/*
 *Modified kdtree.cpp by David Trainor
 *CS 348B: Assignment 2
 *To switch between the original version and the new Lazy Kd-Tree version, comment out #define LAZY_KD_TREE below
 *To switch between fully refining the all second level kd-trees or not, choose the correct #define FULLY_REFINE_SECOND_STEP
 */


/*
 * pbrt source code Copyright(c) 1998-2005 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For educational use only; commercial use expressly forbidden.
 * NO WARRANTY, express or implied, for this software.
 * (See file License.txt for complete license)
 */

// kdtree.cpp*


//Comment this out to use the default version of the Kd-Tree
//#define LAZY_KD_TREE



#ifdef LAZY_KD_TREE

//false: Each time the program finds an object it can refine, it makes a new Kd-Tree
//true: Only make two levels of Kd-Trees, and refine the primitives fully after the first level
//#define FULLY_REFINE_SECOND_STEP		false
#define FULLY_REFINE_SECOND_STEP		true

#include "pbrt.h"
#include "primitive.h"


struct MailboxPrim;
struct LazyKdNodeParameters;
struct KdAccelNode;
struct BoundEdge;

class KdTreeAccel;

struct LazyKdNodeParameters
{
	int nodeNum;
	BBox nodeBounds;
	int* primitiveIndices;
	int numPrims;
	
	int depth;
	int badRefines;
};

struct MailboxPrim
{
	MailboxPrim(const Reference<Primitive> &p)
	{
		primitive = p;
		tree = 0;
		lastMailboxId = -1;
		useTree = false;
		treeBuilt = false;
	}
	MailboxPrim(const Reference<KdTreeAccel> &p)
	{
		primitive = 0;
		tree = p;
		lastMailboxId = -1;
		useTree = true;
		treeBuilt = true;
	}
	MailboxPrim()
	{
		primitive = 0;
		tree = 0;
		lastMailboxId = -1;
		useTree = true;
		treeBuilt = false;
	}
	Reference<Primitive> primitive;
	Reference<KdTreeAccel> tree;
	int lastMailboxId;
	bool useTree;
	bool treeBuilt;
	BBox prim_bound;
};



struct KdAccelNode
{
	void initLeaf(int *primNums, int np, MailboxPrim *mailboxPrims, MemoryArena &arena)
	{
		//printf("initLeaf\n"); fflush(stdout);
		static StatsCounter numLeafMade("Kd-Tree Accelerator", "Leaf kd-tree nodes made");
		static StatsCounter maxLeafPrims("Kd-Tree Accelerator", "Maximum number of primitives in leaf node");
		++numLeafMade;
		maxLeafPrims.Max(np);
		static StatsRatio leafPrims("Kd-Tree Accelerator", "Avg. number of primitives in leaf nodes");
		leafPrims.Add(np, 1);
		nPrims = np << 3;
		//flags |= 3;
		//flags |= 4;
		flags |= 7;
		if (np == 0)
		{
			onePrimitive = NULL;
		} else if (np == 1) {
			onePrimitive = &mailboxPrims[primNums[0]];
		} else {
			primitives = (MailboxPrim **)arena.Alloc(np * sizeof(MailboxPrim *));
			for (int i = 0; i < np; ++i)
			{
				primitives[i] = &mailboxPrims[primNums[i]];
			}
		}
	}

	void initInterior(int axis, float s)
	{
		//printf("initInterior\n"); fflush(stdout);
		static StatsCounter nodesMade("Kd-Tree Accelerator", "Interior kd-tree nodes made"); // NOBOOK
		++nodesMade; // NOBOOK
		split = s;
		flags &= ~3;
		flags |= 4;
		flags |= axis;
	}

	float SplitPos() const { return split; }
	int nPrimitives() const { return nPrims >> 3; }
	int SplitAxis() const { return flags & 3; }
	bool IsLeaf() const { return (flags & 3) == 3; }
	bool IsBuilt() const { return (flags & 4) == 4; }

	union {
		u_int flags;   // Both
		float split;   // Interior
		u_int nPrims;  // Leaf
	};
	
	union {
		u_int aboveChild;           // Interior
		MailboxPrim *onePrimitive;  // Leaf
		MailboxPrim **primitives;   // Leaf
		LazyKdNodeParameters* parameters;	//Not Built
	};
};




struct BoundEdge
{
	BoundEdge()
	{
	}
	
	BoundEdge(float tt, int pn, bool starting)
	{
		t = tt;
		primNum = pn;
		type = starting ? START : END;
	}
	
	bool operator<(const BoundEdge &e) const
	{
		if (t == e.t)
		{
			return (int)type < (int)e.type;
		} else {
			return t < e.t;
		}
	}
	
	float t;
	int primNum;
	enum { START, END } type;
};


class  KdTreeAccel : public Aggregate
{
public:
	KdTreeAccel(const vector<Reference<Primitive> > &p, int icost, int scost, float ebonus, int maxp, int maxDepth, bool refine);
	BBox WorldBound() const { return bounds; }
	bool CanIntersect() const { return true; }
	~KdTreeAccel();

	void buildNode(int nodeNum) const ;

	int allocateNode(const BBox &nodeBounds, int *primNums,
		int nPrims, int depth, int badRefines = 0) const;

	bool Intersect(const Ray &ray, Intersection *isect) const;
	bool IntersectP(const Ray &ray) const;
private:
	// KdTreeAccel Private Data
	int isectCost, traversalCost, maxPrims;
	float emptyBonus;
	u_int nMailboxes;
	bool refinePrims;
	int max_depth;

	MailboxPrim *mailboxPrims;
	mutable int curMailboxId;
	mutable KdAccelNode *nodes;
	mutable int nAllocedNodes, nextFreeNode;
	BBox bounds;
	mutable MemoryArena arena;

	mutable BoundEdge *edges[3];
	vector<BBox> primBounds;
	int *prims0;
	int *prims1;
};


struct KdToDo {
	const KdAccelNode *node;
	float tmin, tmax;
	int nodeNum;
};

struct PrimData
{
	union {
		MailboxPrim* primitive;
		KdTreeAccel* tree;
	};
	bool useTree;
};
KdTreeAccel::KdTreeAccel(const vector<Reference<Primitive> > &p, int icost, int tcost, float ebonus, int maxp, int maxDepth, bool refine)
	: isectCost(icost), traversalCost(tcost), maxPrims(maxp), emptyBonus(ebonus), refinePrims(refine), max_depth(maxDepth)
{
	static StatsCounter treesMade("Kd-Tree Accelerator", "Kd-trees made");
	++treesMade;

	int *primNums;
	curMailboxId = 0;
	if(refinePrims)
	{
		vector<Reference<Primitive > > prims;
		for (u_int i = 0; i < p.size(); ++i)
		{
			p[i]->FullyRefine(prims);
		}
		nMailboxes = prims.size();
		
		mailboxPrims = (MailboxPrim *)AllocAligned(nMailboxes * sizeof(MailboxPrim));
		for(u_int i=0;i<nMailboxes;i++)
		{
			new (&mailboxPrims[i]) MailboxPrim(prims[i]);
		}
		primNums = new int[prims.size()];
		for (u_int i = 0; i < prims.size(); ++i)
		{
			primNums[i] = i;
		}
		
		primBounds.reserve(nMailboxes);
		for (u_int i = 0; i < nMailboxes; ++i)
		{
			BBox b = prims[i]->WorldBound();
			bounds = Union(bounds, b);
			primBounds.push_back(b);
			mailboxPrims[i].prim_bound = b;
		}
	} else {
		nMailboxes = p.size();
		
		mailboxPrims = (MailboxPrim *)AllocAligned(nMailboxes * sizeof(MailboxPrim));
		for(u_int i=0;i<nMailboxes;i++)
		{
			if(p[i]->CanIntersect())
			{
				new (&mailboxPrims[i]) MailboxPrim(p[i]);
			} else {
				new (&mailboxPrims[i]) MailboxPrim();
				mailboxPrims[i].primitive = p[i];
			}
		}
		
		primNums = new int[p.size()];
		for (u_int i = 0; i < p.size(); ++i)
		{
			primNums[i] = i;
		}

		primBounds.reserve(nMailboxes);
		for (u_int i = 0; i < nMailboxes; ++i)
		{
			BBox b = p[i]->WorldBound();
			bounds = Union(bounds, b);
			primBounds.push_back(b);
			mailboxPrims[i].prim_bound = b;
		}
	}

	//printf("--KdTreeAccel-- (%d)\n", nMailboxes); fflush(stdout);
	nextFreeNode = nAllocedNodes = 0;
	if (maxDepth <= 0)
	{
		max_depth = maxDepth = Round2Int(8 + 1.3f * Log2Int(float(nMailboxes)));
	}
	for (int i = 0; i < 3; ++i)
		edges[i] = new BoundEdge[2*nMailboxes];
	prims0 = new int[nMailboxes];
	prims1 = new int[(max_depth+1) * nMailboxes];
	allocateNode(bounds, primNums, nMailboxes, max_depth);
	delete[] primNums;


	static StatsCounter maxPrims("Kd-Tree Accelerator", "Maximum kd-tree size");
	maxPrims.Max(nMailboxes);
	//static StatsCounter minPrims("Kd-Tree Accelerator", "Minimum kd-tree size");
	//minPrims.Min(nMailboxes);
}

KdTreeAccel::~KdTreeAccel()
{
	
	for(int a=0;a<nextFreeNode;a++)
	{
		if(!nodes[a].IsBuilt())
		{
			if(nodes[a].parameters->numPrims > 0)
			{
				delete[] nodes[a].parameters->primitiveIndices;
			}
			delete nodes[a].parameters;
			nodes[a].parameters = 0;
		}
	}
	for (u_int i = 0; i < nMailboxes; ++i)
	{
		if(mailboxPrims[i].useTree && mailboxPrims[i].treeBuilt)
		{
			mailboxPrims[i].tree->~KdTreeAccel();
		}
		mailboxPrims[i].~MailboxPrim();
	}

	FreeAligned(mailboxPrims);
	FreeAligned(nodes);

	for (int i = 0; i < 3; ++i)
		delete[] edges[i];
	
	delete[] prims0;
	delete[] prims1;
	
}

int KdTreeAccel::allocateNode(const BBox &nodeBounds, int *primNums,
		int nPrims, int depth, int badRefines) const
{
	//printf("allocateNode: %d\n", nextFreeNode); fflush(stdout);
	int nodeNum = nextFreeNode;
	if (nextFreeNode == nAllocedNodes)
	{
		int nAlloc = max(2 * nAllocedNodes, 512);
		KdAccelNode *n = (KdAccelNode *)AllocAligned(nAlloc * sizeof(KdAccelNode));
		if (nAllocedNodes > 0)
		{
			memcpy(n, nodes, nAllocedNodes * sizeof(KdAccelNode));
			FreeAligned(nodes);
		}
		nodes = n;
		nAllocedNodes = nAlloc;
	}
	nextFreeNode++;

	LazyKdNodeParameters* parameters;
	parameters = new LazyKdNodeParameters;
	parameters->nodeNum = nodeNum;
	parameters->nodeBounds = nodeBounds;
	parameters->numPrims = nPrims;
	parameters->depth = depth;
	parameters->badRefines = badRefines;
	int* primitiveInds = 0;
	if(nPrims > 0)
	{
		primitiveInds = new int[nPrims];
		memcpy(primitiveInds, primNums, sizeof(int)*nPrims);
	}
	parameters->primitiveIndices = primitiveInds;
	nodes[nodeNum].parameters = parameters;
	nodes[nodeNum].flags &= 0x0000;

	return nodeNum;
}

void KdTreeAccel::buildNode(int nodeNum) const
{
	//printf("buildNode: %d\n", nodeNum); fflush(stdout);
	if(nodeNum > nAllocedNodes)
	{
		printf("building: %d, allocated: %d\n", nodeNum, nAllocedNodes);
		Severe("Trying to build a node that is not allocated yet\n");
		exit(1);
	}
	LazyKdNodeParameters* parameters = nodes[nodeNum].parameters;
	BBox nodeBounds = parameters->nodeBounds;
	int nPrims = parameters->numPrims;
	int depth = parameters->depth;
	int badRefines = parameters->badRefines;
	int* primNums = parameters->primitiveIndices;

	if (nPrims <= maxPrims || depth == 0)
	{
		nodes[nodeNum].initLeaf(primNums, nPrims, mailboxPrims, arena);
		return;
	}
	int bestAxis = -1, bestOffset = -1;
	float bestCost = INFINITY;
	float oldCost = isectCost * float(nPrims);
	
	Vector d = nodeBounds.pMax - nodeBounds.pMin;
	float totalSA = (2.f * (d.x*d.y + d.x*d.z + d.y*d.z));
	float invTotalSA = 1.f / totalSA;
	int axis;
	if (d.x > d.y && d.x > d.z)
		axis = 0;
	else
		axis = (d.y > d.z) ? 1 : 2;

	int retries = 0;
	retrySplit:
	for (int i = 0; i < nPrims; ++i)
	{
		int pn = primNums[i];
		const BBox &bbox = primBounds[pn];
		edges[axis][2*i] = BoundEdge(bbox.pMin[axis], pn, true);
		edges[axis][2*i+1] = BoundEdge(bbox.pMax[axis], pn, false);
	}
	sort(&edges[axis][0], &edges[axis][2*nPrims]);
	int nBelow = 0, nAbove = nPrims;
	for (int i = 0; i < 2*nPrims; ++i)
	{
		if (edges[axis][i].type == BoundEdge::END)
		{
			--nAbove;
		}
		float edget = edges[axis][i].t;
		if (edget > nodeBounds.pMin[axis] && edget < nodeBounds.pMax[axis])
		{
			int otherAxis[3][2] = { {1,2}, {0,2}, {0,1} };
			int otherAxis0 = otherAxis[axis][0];
			int otherAxis1 = otherAxis[axis][1];
			float belowSA = 2 * (d[otherAxis0] * d[otherAxis1] + (edget - nodeBounds.pMin[axis]) * (d[otherAxis0] + d[otherAxis1]));
			float aboveSA = 2 * (d[otherAxis0] * d[otherAxis1] + (nodeBounds.pMax[axis] - edget) * (d[otherAxis0] + d[otherAxis1]));
			float pBelow = belowSA * invTotalSA;
			float pAbove = aboveSA * invTotalSA;
			float eb = (nAbove == 0 || nBelow == 0) ? emptyBonus : 0.f;
			//float cost = traversalCost + isectCost * (1.f - eb) * (pBelow * nBelow + pAbove * nAbove);
			float cost;

			if(!mailboxPrims[((int)i/2)].primitive->CanIntersect())
			{
				float refineMultiple = 1.2f;
				float refineAdd = 0.0f;
				cost = traversalCost + (isectCost * refineMultiple + refineAdd ) * (1.f - eb) * (pBelow * nBelow + pAbove * nAbove);
			} else {
				cost = traversalCost + isectCost * (1.f - eb) * (pBelow * nBelow + pAbove * nAbove);
			}

			if (cost < bestCost)
			{
				bestCost = cost;
				bestAxis = axis;
				bestOffset = i;
			}
		}
		if (edges[axis][i].type == BoundEdge::START)
		{
			++nBelow;
		}
	}
	Assert(nBelow == nPrims && nAbove == 0); // NOBOOK
	if (bestAxis == -1 && retries < 2)
	{
		++retries;
		axis = (axis+1) % 3;
		goto retrySplit;
	}
	if (bestCost > oldCost)
	{
		++badRefines;
	}
	if ((bestCost > 4.f * oldCost && nPrims < 16) || bestAxis == -1 || badRefines == 3)
	{
		nodes[nodeNum].initLeaf(primNums, nPrims, mailboxPrims, arena);
		return;
	}
	int n0 = 0, n1 = 0;
	for (int i = 0; i < bestOffset; ++i)
	{
		if (edges[bestAxis][i].type == BoundEdge::START)
		{
			prims0[n0++] = edges[bestAxis][i].primNum;
		}
	}
	for (int i = bestOffset+1; i < 2*nPrims; ++i)
	{
		if (edges[bestAxis][i].type == BoundEdge::END)
		{
			prims1[n1++] = edges[bestAxis][i].primNum;
		}
	}
	float tsplit = edges[bestAxis][bestOffset].t;
	nodes[nodeNum].initInterior(bestAxis, tsplit);

	BBox bounds0 = nodeBounds;
	BBox bounds1 = nodeBounds;
	bounds0.pMax[bestAxis] = bounds1.pMin[bestAxis] = tsplit;

	//nodes[nodeNum].aboveChild = nextFreeNode;
	
	nodes[nodeNum].aboveChild = allocateNode(bounds0, prims0, n0, depth-1, badRefines);
	int tmp = allocateNode(bounds1, prims1, n1, depth-1, badRefines);

	if(tmp != nodes[nodeNum].aboveChild + 1)
	{
		Severe("nodes not adjacent!\n");
		exit(1);
	}
	/*
	if(nodeNum == 32766)
	{
		printf("\nChildren ptrs are: %d, %d!\n", nodes[nodeNum].aboveChild, nodes[nodeNum].aboveChild+1);
		printf("To check, child thinks it is: %d\n", nodes[nodes[nodeNum].aboveChild].parameters->nodeNum);
	}
	*/
	if(parameters->numPrims > 0)
	{
		delete[] parameters->primitiveIndices;
	}
	delete parameters;
	/*
	if(nodes[nodeNum].aboveChild == 65536 || nodes[nodeNum].aboveChild+1 == 65536)
	{
		printf("parent is: %d\n", nodeNum);
		fflush(stdout);
	}
	*/
}


bool KdTreeAccel::Intersect(const Ray &ray, Intersection *isect) const
{
	//printf("--Intersect--\n"); fflush(stdout);
	float tmin, tmax;
	if (!bounds.IntersectP(ray, &tmin, &tmax))
	{
		return false;
	}
	int rayId = curMailboxId++;
	Vector invDir(1.f/ray.d.x, 1.f/ray.d.y, 1.f/ray.d.z);
	#define MAX_TODO 64
	KdToDo todo[MAX_TODO];
	int todoPos = 0;
	bool hit = false;
	const KdAccelNode *node = &nodes[0];

	int curNodeNum = 0;
	while (node != NULL)
	{
		if (ray.maxt < tmin)
			break;
		//static StatsCounter nodesTraversed("Kd-Tree Accelerator", //NOBOOK
		//	"Number of kd-tree nodes traversed by normal rays"); //NOBOOK
		//++nodesTraversed; //NOBOOK

		node = &nodes[curNodeNum];
		if(!node->IsBuilt())
		{
			buildNode(node->parameters->nodeNum);
			node = &nodes[curNodeNum];
		}
		if (!node->IsLeaf())
		{
			int axis = node->SplitAxis();
			float tplane = (node->SplitPos() - ray.o[axis]) * invDir[axis];
			const KdAccelNode *firstChild, *secondChild;
			int belowFirst = ray.o[axis] <= node->SplitPos();
			if (belowFirst)
			{
				firstChild = &nodes[node->aboveChild];
				secondChild = &nodes[node->aboveChild+1];
			} else {
				firstChild = &nodes[node->aboveChild+1];
				secondChild = &nodes[node->aboveChild];
			}
			if (tplane > tmax || tplane < 0)
			{
				if(belowFirst)
				{
					curNodeNum = node->aboveChild;
				} else {
					curNodeNum = node->aboveChild+1;
				}
				node = firstChild;
			} else if (tplane < tmin) {
				if(belowFirst)
				{
					curNodeNum = node->aboveChild+1;
				} else {
					curNodeNum = node->aboveChild;
				}
				node = secondChild;
			} else {
				todo[todoPos].node = secondChild;
				todo[todoPos].tmin = tplane;
				todo[todoPos].tmax = tmax;
				if(belowFirst)
				{
					todo[todoPos].nodeNum = node->aboveChild+1;
					curNodeNum = node->aboveChild;
				} else {
					todo[todoPos].nodeNum = node->aboveChild;
					curNodeNum = node->aboveChild+1;
				}
				++todoPos;
				node = firstChild;
				tmax = tplane;
			}
		} else {
			u_int nPrimitives = node->nPrimitives();
			if (nPrimitives == 1)
			{
				MailboxPrim *mp = node->onePrimitive;
				if (mp->lastMailboxId != rayId)
				{
					mp->lastMailboxId = rayId;
					if(mp->useTree)
					{
						if(!mp->treeBuilt)
						{
							if(mp->prim_bound.IntersectP(ray))
							{
								if(FULLY_REFINE_SECOND_STEP)
								{
									vector<Reference<Primitive>> prim_tmp;
									prim_tmp.push_back(mp->primitive);
									mp->tree = new KdTreeAccel(prim_tmp, isectCost, traversalCost, emptyBonus, 1, -1, true);
								} else {
									vector<Reference<Primitive>> prim_tmp;
									mp->primitive->Refine(prim_tmp);
									mp->tree = new KdTreeAccel(prim_tmp, isectCost, traversalCost, emptyBonus, 1, -1, false);
								}
								mp->treeBuilt = true;
							}
						}
						if(mp->treeBuilt)
						{
							if(mp->tree->Intersect(ray, isect))
								hit = true;
						}
					} else {
						if (mp->primitive->Intersect(ray, isect))
							hit = true;
					}
				}
			} else {
				MailboxPrim **prims = node->primitives;
				for (u_int i = 0; i < nPrimitives; ++i)
				{
					MailboxPrim *mp = prims[i];
					if (mp->lastMailboxId != rayId)
					{
						mp->lastMailboxId = rayId;
						if(mp->useTree)
						{
							if(!mp->treeBuilt)
							{
								if(mp->prim_bound.IntersectP(ray))
								{
									if(FULLY_REFINE_SECOND_STEP)
									{
										vector<Reference<Primitive>> prim_tmp;
										prim_tmp.push_back(mp->primitive);
										mp->tree = new KdTreeAccel(prim_tmp, isectCost, traversalCost, emptyBonus, 1, -1, true);
									} else {
										vector<Reference<Primitive>> prim_tmp;
										mp->primitive->Refine(prim_tmp);
										mp->tree = new KdTreeAccel(prim_tmp, isectCost, traversalCost, emptyBonus, 1, -1, false);
									}
									mp->treeBuilt = true;
								}
							}
							if(mp->treeBuilt)
							{
								if(mp->tree->Intersect(ray, isect))
									hit = true;
							}
						} else {
							if (mp->primitive->Intersect(ray, isect))
								hit = true;
						}
					}
				}
			}
			if (todoPos > 0)
			{
				--todoPos;
				node = todo[todoPos].node;
				tmin = todo[todoPos].tmin;
				tmax = todo[todoPos].tmax;
				curNodeNum = todo[todoPos].nodeNum;
			} else
				break;
		}
	}
	return hit;
}


bool KdTreeAccel::IntersectP(const Ray &ray) const
{
	//printf("--IntersectP--\n"); fflush(stdout);
	float tmin, tmax;
	if (!bounds.IntersectP(ray, &tmin, &tmax))
		return false;
	int rayId = curMailboxId++;
	Vector invDir(1.f/ray.d.x, 1.f/ray.d.y, 1.f/ray.d.z);
	#define MAX_TODO 64
	KdToDo todo[MAX_TODO];
	int todoPos = 0;
	const KdAccelNode *node = &nodes[0];
	

	int curNodeNum = 0;
	while (node != NULL)
	{
		// Update kd-tree shadow ray traversal statistics
		//static StatsCounter nodesTraversed("Kd-Tree Accelerator",
		//	"Number of kd-tree nodes traversed by shadow rays");
		//++nodesTraversed;
		
		node = &nodes[curNodeNum];
		if(!node->IsBuilt())
		{
			buildNode(node->parameters->nodeNum);
			node = &nodes[curNodeNum];
		}
		if (node->IsLeaf())
		{
			u_int nPrimitives = node->nPrimitives();
			if (nPrimitives == 1) {
				MailboxPrim *mp = node->onePrimitive;
				if (mp->lastMailboxId != rayId) {
					mp->lastMailboxId = rayId;

					if(mp->useTree)
					{
						if(!mp->treeBuilt)
						{
							if(mp->prim_bound.IntersectP(ray))
							{
								if(FULLY_REFINE_SECOND_STEP)
								{
									vector<Reference<Primitive>> prim_tmp;
									prim_tmp.push_back(mp->primitive);
									mp->tree = new KdTreeAccel(prim_tmp, isectCost, traversalCost, emptyBonus, 1, -1, true);
								} else {
									vector<Reference<Primitive>> prim_tmp;
									mp->primitive->Refine(prim_tmp);
									mp->tree = new KdTreeAccel(prim_tmp, isectCost, traversalCost, emptyBonus, 1, -1, false);
								}
								mp->treeBuilt = true;
							}
						}
						if(mp->treeBuilt)
						{
							if(mp->tree->IntersectP(ray))
								return true;
						}
					} else {
						if (mp->primitive->IntersectP(ray))
							return true;
					}
				}
			} else {
				MailboxPrim **prims = node->primitives;
				for (u_int i = 0; i < nPrimitives; ++i) {
					MailboxPrim *mp = prims[i];
					if (mp->lastMailboxId != rayId) {
						mp->lastMailboxId = rayId;

						if(mp->useTree)
						{
							if(!mp->treeBuilt)
							{
								if(mp->prim_bound.IntersectP(ray))
								{
									if(FULLY_REFINE_SECOND_STEP)
									{
										vector<Reference<Primitive>> prim_tmp;
										prim_tmp.push_back(mp->primitive);
										mp->tree = new KdTreeAccel(prim_tmp, isectCost, traversalCost, emptyBonus, 1, -1, true);
									} else {
										vector<Reference<Primitive>> prim_tmp;
										mp->primitive->Refine(prim_tmp);
										mp->tree = new KdTreeAccel(prim_tmp, isectCost, traversalCost, emptyBonus, 1, -1, false);
									}
									mp->treeBuilt = true;
								}
							}
							if(mp->treeBuilt)
							{
								if(mp->tree->IntersectP(ray))
									return true;
							}
						} else {
							if (mp->primitive->IntersectP(ray))
								return true;
						}
					}
				}
			}
			if (todoPos > 0)
			{
				--todoPos;
				node = todo[todoPos].node;
				tmin = todo[todoPos].tmin;
				tmax = todo[todoPos].tmax;
				curNodeNum = todo[todoPos].nodeNum;

			} else
				break;
		} else {
			int axis = node->SplitAxis();
			float tplane = (node->SplitPos() - ray.o[axis]) * invDir[axis];
			const KdAccelNode *firstChild, *secondChild;
			int belowFirst = ray.o[axis] <= node->SplitPos();
			if (belowFirst)
			{
				firstChild = &nodes[node->aboveChild];
				secondChild = &nodes[node->aboveChild+1];
			} else {
				firstChild = &nodes[node->aboveChild+1];
				secondChild = &nodes[node->aboveChild];
			}
			
			if (tplane > tmax || tplane < 0)
			{
				if(belowFirst)
				{
					curNodeNum = node->aboveChild;
				} else {
					curNodeNum = node->aboveChild+1;
				}
				node = firstChild;
			} else if (tplane < tmin)
			{
				if(belowFirst)
				{
					curNodeNum = node->aboveChild+1;
				} else {
					curNodeNum = node->aboveChild;
				}
				node = secondChild;
			} else {
				todo[todoPos].node = secondChild;
				todo[todoPos].tmin = tplane;
				todo[todoPos].tmax = tmax;
				if(belowFirst)
				{
					todo[todoPos].nodeNum = node->aboveChild+1;
					curNodeNum = node->aboveChild;
				} else {
					todo[todoPos].nodeNum = node->aboveChild;
					curNodeNum = node->aboveChild+1;
				}
				++todoPos;
				node = firstChild;
				tmax = tplane;
			}
		}
	}
	return false;
}


extern "C" DLLEXPORT Primitive *CreateAccelerator(const vector<Reference<Primitive> > &prims,
		const ParamSet &ps) {
	int isectCost = ps.FindOneInt("intersectcost", 80);
	int travCost = ps.FindOneInt("traversalcost", 1);
	float emptyBonus = ps.FindOneFloat("emptybonus", 0.5f);
	int maxPrims = ps.FindOneInt("maxprims", 1);
	int maxDepth = ps.FindOneInt("maxdepth", -1);
	ProgressReporter progress(1, "Building KDTree");
	KdTreeAccel* accel = new KdTreeAccel(prims, isectCost, travCost, emptyBonus, maxPrims, maxDepth, false);
	progress.Update();
	progress.Done();
	return accel;
}





#else
//OLD CODE



#include "pbrt.h"
#include "primitive.h"
// KdAccelNode Declarations
struct MailboxPrim {
	MailboxPrim(const Reference<Primitive> &p) {
		primitive = p;
		lastMailboxId = -1;
	}
	Reference<Primitive> primitive;
	int lastMailboxId;
};
struct KdAccelNode {
	// KdAccelNode Methods
	void initLeaf(int *primNums, int np,
			MailboxPrim *mailboxPrims, MemoryArena &arena) {
		// Update kd leaf node allocation statistics
		static StatsCounter numLeafMade("Kd-Tree Accelerator",
		                                "Leaf kd-tree nodes made");
		//static StatsCounter maxDepth("Kd-Tree Accelerator",
		//                             "Maximum kd-tree depth");
		static StatsCounter maxLeafPrims("Kd-Tree Accelerator",
			"Maximum number of primitives in leaf node");
		++numLeafMade;
		//maxDepth.Max(depth);
		maxLeafPrims.Max(np);
		static StatsRatio leafPrims("Kd-Tree Accelerator",
			"Avg. number of primitives in leaf nodes");
		leafPrims.Add(np, 1);
		nPrims = np << 2;
		flags |= 3;
		// Store _MailboxPrim *_s for leaf node
		if (np == 0)
			onePrimitive = NULL;
		else if (np == 1)
			onePrimitive = &mailboxPrims[primNums[0]];
		else {
			primitives = (MailboxPrim **)arena.Alloc(np *
				sizeof(MailboxPrim *));
			for (int i = 0; i < np; ++i)
				primitives[i] = &mailboxPrims[primNums[i]];
		}
	}
	void initInterior(int axis, float s) {
		static StatsCounter nodesMade("Kd-Tree Accelerator", "Interior kd-tree nodes made"); // NOBOOK
		++nodesMade; // NOBOOK
		split = s;
		flags &= ~3;
		flags |= axis;
	}
	float SplitPos() const { return split; }
	int nPrimitives() const { return nPrims >> 2; }
	int SplitAxis() const { return flags & 3; }
	bool IsLeaf() const { return (flags & 3) == 3; }
	union {
		u_int flags;   // Both
		float split;   // Interior
		u_int nPrims;  // Leaf
	};
	union {
		u_int aboveChild;           // Interior
		MailboxPrim *onePrimitive;  // Leaf
		MailboxPrim **primitives;   // Leaf
	};
};
struct BoundEdge {
	// BoundEdge Public Methods
	BoundEdge() { }
	BoundEdge(float tt, int pn, bool starting) {
		t = tt;
		primNum = pn;
		type = starting ? START : END;
	}
	bool operator<(const BoundEdge &e) const {
		if (t == e.t)
			return (int)type < (int)e.type;
		else return t < e.t;
	}
	float t;
	int primNum;
	enum { START, END } type;
};
// KdTreeAccel Declarations
struct KdAccelNode;
class  KdTreeAccel : public Aggregate {
public:
	// KdTreeAccel Public Methods
	KdTreeAccel(const vector<Reference<Primitive> > &p,
		int icost, int scost,
		float ebonus, int maxp, int maxDepth);
	BBox WorldBound() const { return bounds; }
	bool CanIntersect() const { return true; }
	~KdTreeAccel();
	void buildTree(int nodeNum, const BBox &bounds,
	    const vector<BBox> &primBounds,
		int *primNums, int nprims, int depth,
		BoundEdge *edges[3],
		int *prims0, int *prims1, int badRefines = 0);
	bool Intersect(const Ray &ray, Intersection *isect) const;
	bool IntersectP(const Ray &ray) const;
private:
	// KdTreeAccel Private Data
	int isectCost, traversalCost, maxPrims;
	float emptyBonus;
	u_int nMailboxes;
	MailboxPrim *mailboxPrims;
	mutable int curMailboxId;
	KdAccelNode *nodes;
	int nAllocedNodes, nextFreeNode;
	BBox bounds;
	MemoryArena arena;
};
struct KdToDo {
	const KdAccelNode *node;
	float tmin, tmax;
};
// KdTreeAccel Method Definitions
KdTreeAccel::
    KdTreeAccel(const vector<Reference<Primitive> > &p,
		int icost, int tcost,
		float ebonus, int maxp, int maxDepth)
	: isectCost(icost), traversalCost(tcost),
	maxPrims(maxp), emptyBonus(ebonus) {
	vector<Reference<Primitive > > prims;
	for (u_int i = 0; i < p.size(); ++i)
		p[i]->FullyRefine(prims);
	// Initialize mailboxes for _KdTreeAccel_
	curMailboxId = 0;
	nMailboxes = prims.size();
	mailboxPrims = (MailboxPrim *)AllocAligned(nMailboxes *
		sizeof(MailboxPrim));
	for (u_int i = 0; i < nMailboxes; ++i)
		new (&mailboxPrims[i]) MailboxPrim(prims[i]);
	// Build kd-tree for accelerator
	nextFreeNode = nAllocedNodes = 0;
	if (maxDepth <= 0)
		maxDepth =
		    Round2Int(8 + 1.3f * Log2Int(float(prims.size())));
	// Compute bounds for kd-tree construction
	vector<BBox> primBounds;
	primBounds.reserve(prims.size());
	for (u_int i = 0; i < prims.size(); ++i) {
		BBox b = prims[i]->WorldBound();
		bounds = Union(bounds, b);
		primBounds.push_back(b);
	}
	// Allocate working memory for kd-tree construction
	BoundEdge *edges[3];
	for (int i = 0; i < 3; ++i)
		edges[i] = new BoundEdge[2*prims.size()];
	int *prims0 = new int[prims.size()];
	int *prims1 = new int[(maxDepth+1) * prims.size()];
	// Initialize _primNums_ for kd-tree construction
	int *primNums = new int[prims.size()];
	for (u_int i = 0; i < prims.size(); ++i)
		primNums[i] = i;
	// Start recursive construction of kd-tree
	buildTree(0, bounds, primBounds, primNums,
	          prims.size(), maxDepth, edges,
			  prims0, prims1);
	// Free working memory for kd-tree construction
	delete[] primNums;
	for (int i = 0; i < 3; ++i)
		delete[] edges[i];
	delete[] prims0;
	delete[] prims1;
}
KdTreeAccel::~KdTreeAccel() {
	for (u_int i = 0; i < nMailboxes; ++i)
		mailboxPrims[i].~MailboxPrim();
	FreeAligned(mailboxPrims);
	FreeAligned(nodes);
}
void KdTreeAccel::buildTree(int nodeNum,
        const BBox &nodeBounds,
		const vector<BBox> &allPrimBounds, int *primNums,
		int nPrims, int depth, BoundEdge *edges[3],
		int *prims0, int *prims1, int badRefines) {
	Assert(nodeNum == nextFreeNode); // NOBOOK
	// Get next free node from _nodes_ array
	if (nextFreeNode == nAllocedNodes) {
		int nAlloc = max(2 * nAllocedNodes, 512);
		KdAccelNode *n = (KdAccelNode *)AllocAligned(nAlloc *
			sizeof(KdAccelNode));
		if (nAllocedNodes > 0) {
			memcpy(n, nodes,
			       nAllocedNodes * sizeof(KdAccelNode));
			FreeAligned(nodes);
		}
		nodes = n;
		nAllocedNodes = nAlloc;
	}
	++nextFreeNode;
	// Initialize leaf node if termination criteria met
	if (nPrims <= maxPrims || depth == 0) {
		nodes[nodeNum].initLeaf(primNums, nPrims,
		                       mailboxPrims, arena);
		return;
	}
	// Initialize interior node and continue recursion
	// Choose split axis position for interior node
	int bestAxis = -1, bestOffset = -1;
	float bestCost = INFINITY;
	float oldCost = isectCost * float(nPrims);
	Vector d = nodeBounds.pMax - nodeBounds.pMin;
	float totalSA = (2.f * (d.x*d.y + d.x*d.z + d.y*d.z));
	float invTotalSA = 1.f / totalSA;
	// Choose which axis to split along
	int axis;
	if (d.x > d.y && d.x > d.z) axis = 0;
	else axis = (d.y > d.z) ? 1 : 2;
	int retries = 0;
	retrySplit:
	// Initialize edges for _axis_
	for (int i = 0; i < nPrims; ++i) {
		int pn = primNums[i];
		const BBox &bbox = allPrimBounds[pn];
		edges[axis][2*i] =
		    BoundEdge(bbox.pMin[axis], pn, true);
		edges[axis][2*i+1] =
			BoundEdge(bbox.pMax[axis], pn, false);
	}
	sort(&edges[axis][0], &edges[axis][2*nPrims]);
	// Compute cost of all splits for _axis_ to find best
	int nBelow = 0, nAbove = nPrims;
	for (int i = 0; i < 2*nPrims; ++i) {
		if (edges[axis][i].type == BoundEdge::END) --nAbove;
		float edget = edges[axis][i].t;
		if (edget > nodeBounds.pMin[axis] &&
			edget < nodeBounds.pMax[axis]) {
			// Compute cost for split at _i_th edge
			int otherAxis[3][2] = { {1,2}, {0,2}, {0,1} };
			int otherAxis0 = otherAxis[axis][0];
			int otherAxis1 = otherAxis[axis][1];
			float belowSA = 2 * (d[otherAxis0] * d[otherAxis1] +
			             		(edget - nodeBounds.pMin[axis]) *
				                (d[otherAxis0] + d[otherAxis1]));
			float aboveSA = 2 * (d[otherAxis0] * d[otherAxis1] +
								(nodeBounds.pMax[axis] - edget) *
								(d[otherAxis0] + d[otherAxis1]));
			float pBelow = belowSA * invTotalSA;
			float pAbove = aboveSA * invTotalSA;
			float eb = (nAbove == 0 || nBelow == 0) ? emptyBonus : 0.f;
			float cost = traversalCost + isectCost * (1.f - eb) *
				(pBelow * nBelow + pAbove * nAbove);
			// Update best split if this is lowest cost so far
			if (cost < bestCost)  {
				bestCost = cost;
				bestAxis = axis;
				bestOffset = i;
			}
		}
		if (edges[axis][i].type == BoundEdge::START) ++nBelow;
	}
	Assert(nBelow == nPrims && nAbove == 0); // NOBOOK
	// Create leaf if no good splits were found
	if (bestAxis == -1 && retries < 2) {
		++retries;
		axis = (axis+1) % 3;
		goto retrySplit;
	}
	if (bestCost > oldCost) ++badRefines;
	if ((bestCost > 4.f * oldCost && nPrims < 16) ||
		bestAxis == -1 || badRefines == 3) {
		nodes[nodeNum].initLeaf(primNums, nPrims,
		                     mailboxPrims, arena);
		return;
	}
	// Classify primitives with respect to split
	int n0 = 0, n1 = 0;
	for (int i = 0; i < bestOffset; ++i)
		if (edges[bestAxis][i].type == BoundEdge::START)
			prims0[n0++] = edges[bestAxis][i].primNum;
	for (int i = bestOffset+1; i < 2*nPrims; ++i)
		if (edges[bestAxis][i].type == BoundEdge::END)
			prims1[n1++] = edges[bestAxis][i].primNum;
	// Recursively initialize children nodes
	float tsplit = edges[bestAxis][bestOffset].t;
	nodes[nodeNum].initInterior(bestAxis, tsplit);
	BBox bounds0 = nodeBounds, bounds1 = nodeBounds;
	bounds0.pMax[bestAxis] = bounds1.pMin[bestAxis] = tsplit;
	buildTree(nodeNum+1, bounds0,
		allPrimBounds, prims0, n0, depth-1, edges,
		prims0, prims1 + nPrims, badRefines);
	nodes[nodeNum].aboveChild = nextFreeNode;
	buildTree(nodes[nodeNum].aboveChild, bounds1, allPrimBounds,
		prims1, n1, depth-1, edges,
		prims0, prims1 + nPrims, badRefines);
}
bool KdTreeAccel::Intersect(const Ray &ray,
		Intersection *isect) const {
	// Compute initial parametric range of ray inside kd-tree extent
	float tmin, tmax;
	if (!bounds.IntersectP(ray, &tmin, &tmax))
		return false;
	// Prepare to traverse kd-tree for ray
	int rayId = curMailboxId++;
	Vector invDir(1.f/ray.d.x, 1.f/ray.d.y, 1.f/ray.d.z);
	#define MAX_TODO 64
	KdToDo todo[MAX_TODO];
	int todoPos = 0;
	// Traverse kd-tree nodes in order for ray
	bool hit = false;
	const KdAccelNode *node = &nodes[0];
	while (node != NULL) {
		// Bail out if we found a hit closer than the current node
		//printf("Comparing %f to %f\n", ray.maxt, tmin);
		if (ray.maxt < tmin)
		{
			//printf("max t, breaking\n");
			break;
		}
		//static StatsCounter nodesTraversed("Kd-Tree Accelerator", //NOBOOK
		//	"Number of kd-tree nodes traversed by normal rays"); //NOBOOK
		//++nodesTraversed; //NOBOOK
		if (!node->IsLeaf()) {
			// Process kd-tree interior node
			// Compute parametric distance along ray to split plane
			int axis = node->SplitAxis();
			float tplane = (node->SplitPos() - ray.o[axis]) *
				invDir[axis];
			// Get node children pointers for ray
			const KdAccelNode *firstChild, *secondChild;
			int belowFirst = ray.o[axis] <= node->SplitPos();
			if (belowFirst) {
				firstChild = node + 1;
				secondChild = &nodes[node->aboveChild];
			}
			else {
				firstChild = &nodes[node->aboveChild];
				secondChild = node + 1;
			}
			// Advance to next child node, possibly enqueue other child
			if (tplane > tmax || tplane < 0)
				node = firstChild;
			else if (tplane < tmin)
				node = secondChild;
			else {
				// Enqueue _secondChild_ in todo list
				todo[todoPos].node = secondChild;
				todo[todoPos].tmin = tplane;
				todo[todoPos].tmax = tmax;
				++todoPos;
				node = firstChild;
				tmax = tplane;
			}
		}
		else {
			// Check for intersections inside leaf node
			u_int nPrimitives = node->nPrimitives();
			if (nPrimitives == 1) {
				MailboxPrim *mp = node->onePrimitive;
				// Check one primitive inside leaf node
				if (mp->lastMailboxId != rayId) {
					mp->lastMailboxId = rayId;
					if (mp->primitive->Intersect(ray, isect))
						hit = true;
				}
			}
			else {
				MailboxPrim **prims = node->primitives;
				for (u_int i = 0; i < nPrimitives; ++i) {
					MailboxPrim *mp = prims[i];
					// Check one primitive inside leaf node
					if (mp->lastMailboxId != rayId) {
						mp->lastMailboxId = rayId;
						if (mp->primitive->Intersect(ray, isect))
							hit = true;
					}
				}
			}
			// Grab next node to process from todo list
			if (todoPos > 0) {
				--todoPos;
				node = todo[todoPos].node;
				tmin = todo[todoPos].tmin;
				tmax = todo[todoPos].tmax;
			}
			else
				break;
		}
	}
	return hit;
}
bool KdTreeAccel::IntersectP(const Ray &ray) const {
	// Compute initial parametric range of ray inside kd-tree extent
	float tmin, tmax;
	if (!bounds.IntersectP(ray, &tmin, &tmax))
		return false;
	// Prepare to traverse kd-tree for ray
	int rayId = curMailboxId++;
	Vector invDir(1.f/ray.d.x, 1.f/ray.d.y, 1.f/ray.d.z);
	#define MAX_TODO 64
	KdToDo todo[MAX_TODO];
	int todoPos = 0;
	const KdAccelNode *node = &nodes[0];
	while (node != NULL) {
		// Update kd-tree shadow ray traversal statistics
		//static StatsCounter nodesTraversed("Kd-Tree Accelerator",
		//	"Number of kd-tree nodes traversed by shadow rays");
		//++nodesTraversed;
		if (node->IsLeaf()) {
			// Check for shadow ray intersections inside leaf node
			u_int nPrimitives = node->nPrimitives();
			if (nPrimitives == 1) {
				MailboxPrim *mp = node->onePrimitive;
				if (mp->lastMailboxId != rayId) {
					mp->lastMailboxId = rayId;
					if (mp->primitive->IntersectP(ray))
						return true;
				}
			}
			else {
				MailboxPrim **prims = node->primitives;
				for (u_int i = 0; i < nPrimitives; ++i) {
					MailboxPrim *mp = prims[i];
					if (mp->lastMailboxId != rayId) {
						mp->lastMailboxId = rayId;
						if (mp->primitive->IntersectP(ray))
							return true;
					}
				}
			}
			// Grab next node to process from todo list
			if (todoPos > 0) {
				--todoPos;
				node = todo[todoPos].node;
				tmin = todo[todoPos].tmin;
				tmax = todo[todoPos].tmax;
			}
			else
				break;
		}
		else {
			// Process kd-tree interior node
			// Compute parametric distance along ray to split plane
			int axis = node->SplitAxis();
			float tplane = (node->SplitPos() - ray.o[axis]) *
				invDir[axis];
			// Get node children pointers for ray
			const KdAccelNode *firstChild, *secondChild;
			int belowFirst = ray.o[axis] <= node->SplitPos();
			if (belowFirst) {
				firstChild = node + 1;
				secondChild = &nodes[node->aboveChild];
			}
			else {
				firstChild = &nodes[node->aboveChild];
				secondChild = node + 1;
			}
			// Advance to next child node, possibly enqueue other child
			if (tplane > tmax || tplane < 0)
				node = firstChild;
			else if (tplane < tmin)
				node = secondChild;
			else {
				// Enqueue _secondChild_ in todo list
				todo[todoPos].node = secondChild;
				todo[todoPos].tmin = tplane;
				todo[todoPos].tmax = tmax;
				++todoPos;
				node = firstChild;
				tmax = tplane;
			}
		}
	}
	return false;
}
extern "C" DLLEXPORT Primitive *CreateAccelerator(const vector<Reference<Primitive> > &prims, const ParamSet &ps) {
	int isectCost = ps.FindOneInt("intersectcost", 80);
	int travCost = ps.FindOneInt("traversalcost", 1);
	float emptyBonus = ps.FindOneFloat("emptybonus", 0.5f);
	int maxPrims = ps.FindOneInt("maxprims", 1);
	int maxDepth = ps.FindOneInt("maxdepth", -1);
	ProgressReporter progress(1, "Building KDTree");
	KdTreeAccel *accel = new KdTreeAccel(prims, isectCost, travCost,
		emptyBonus, maxPrims, maxDepth);
	progress.Update();
	progress.Done();
	return accel;
}



#endif