
/*
 * pbrt source code Copyright(c) 1998-2005 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For educational use only; commercial use expressly forbidden.
 * NO WARRANTY, express or implied, for this software.
 * (See file License.txt for complete license)
 */

// photonmap.cpp*
#include "pbrt.h"
#include "transport.h"
#include "scene.h"
#include "mc.h"
#include "kdtree.h"
#include "sampling.h"
#include "photonmap.h"


#ifndef COMBINE_WITH_PHOTON_SURFACE
// Photonmap Local Declarations
struct Photon;
struct ClosePhoton;
struct PhotonProcess;
class PhotonIntegrator : public SurfaceIntegrator {
public:
	// PhotonIntegrator Public Methods
	PhotonIntegrator(int ncaus, int ndir, int nindir, int nLookup, int mdepth,
		float maxdist, bool finalGather, int gatherSamples,
		bool directWithPhotons, bool ignoreFirst);

	~PhotonIntegrator();

	Spectrum Li(const Scene *scene, const RayDifferential &ray,
		const Sample *sample, float *alpha) const;

	void RequestSamples(Sample *sample, const Scene *scene);

	void Preprocess(const Scene *);

private:
	// PhotonIntegrator Private Methods
	static inline bool unsuccessful(int needed, int found, int shot) {
		return (found < needed &&
		        (found == 0 || found < shot / 1024));
	}

	static Spectrum LPhoton(KdTree<Photon, PhotonProcess> *map,
		int nPaths, int nLookup, BSDF *bsdf, const Intersection &isect,
		const Vector &w, float maxDistSquared);

	// PhotonIntegrator Private Data
	u_int nCausticPhotons, nIndirectPhotons, nDirectPhotons;
	u_int nLookup;
	mutable int specularDepth;
	int maxSpecularDepth;
	float maxDistSquared;
	bool directWithPhotons, finalGather;
	int gatherSamples;
	// Declare sample parameters for light source sampling
	int *lightSampleOffset, lightNumOffset;
	int *bsdfSampleOffset, *bsdfComponentOffset;
	int gatherSampleOffset, gatherComponentOffset;
	int nCausticPaths, nDirectPaths, nIndirectPaths;
	mutable KdTree<Photon, PhotonProcess> *causticMap;
	mutable KdTree<Photon, PhotonProcess> *directMap;
	mutable KdTree<Photon, PhotonProcess> *indirectMap;

	bool ignoreFirstLight;
};

struct Photon {
	// Photon Constructor
	Photon(const Point &pp, const Spectrum &wt, const Vector &w)
		: p(pp), alpha(wt), wi(w) {
	}

	Photon() { }

	Point p;
	Spectrum alpha;
	Vector wi;
};

struct PhotonProcess {
	// PhotonProcess Public Methods
	PhotonProcess(u_int mp, const Point &p);

	void operator()(const Photon &photon, float dist2, float &maxDistSquared) const;

	const Point &p;
	ClosePhoton *photons;
	u_int nLookup;
	mutable u_int foundPhotons;
};

struct ClosePhoton {
	ClosePhoton(const Photon *p = NULL,
	            float md2 = INFINITY) {
		photon = p;
		distanceSquared = md2;
	}

	bool operator<(const ClosePhoton &p2) const {
		return distanceSquared == p2.distanceSquared ? (photon < p2.photon) :
			distanceSquared < p2.distanceSquared;
	}

	const Photon *photon;
	float distanceSquared;
};
#endif

// Photonmap Method Definitions
PhotonIntegrator::PhotonIntegrator(int ncaus, int ndir, int nind,
		int nl,	int mdepth, float mdist, bool fg,
		int gs, bool dp, bool ignoreFirst)
{
	nCausticPhotons = ncaus;
	nIndirectPhotons = nind;
	nDirectPhotons = ndir;
	nLookup = nl;
	maxDistSquared = mdist * mdist;
	maxSpecularDepth = mdepth;
	causticMap = directMap = indirectMap = NULL;
	specularDepth = 0;
	finalGather = fg;
	gatherSamples = gs;
	directWithPhotons = dp;

	ignoreFirstLight = ignoreFirst;
#ifdef COMBINE_WITH_PHOTON_SURFACE
	nIndirectPathsSoFar = 0;
#endif
}

PhotonIntegrator::~PhotonIntegrator()
{
	delete causticMap;
	delete directMap;
	delete indirectMap;
}

void PhotonIntegrator::RequestSamples(Sample *sample,
		const Scene *scene)
{
	// Allocate and request samples for sampling all lights
	
	//grab the number of lights
	u_int nLights = scene->lights.size();

	//create arrays for light offsets
	lightSampleOffset = new int[nLights];
	bsdfSampleOffset = new int[nLights];
	bsdfComponentOffset = new int[nLights];
	
	//for each light
	for (u_int i = 0; i < nLights; ++i) {
		//grab the current light
		const Light *light = scene->lights[i];

		//how many samples for this light?
		int lightSamples =
			scene->sampler->RoundSize(light->nSamples);
		
		//add the number of samples
		lightSampleOffset[i] = sample->Add2D(lightSamples);
		bsdfSampleOffset[i] = sample->Add2D(lightSamples);
		bsdfComponentOffset[i] = sample->Add1D(lightSamples);
	}

	//set the light num offset to -1
	lightNumOffset = -1;

	//if we're final gathering...
	if (finalGather) {
		//do roundsize on the gather samples
		gatherSamples = scene->sampler->RoundSize(gatherSamples);
		//grab the offset?
		gatherSampleOffset = sample->Add2D(gatherSamples);
		//grab the component offset?
		gatherComponentOffset = sample->Add1D(gatherSamples);
	}
}

void PhotonIntegrator::Preprocess(const Scene *scene) {

	//do we even have any lights?
	if (scene->lights.size() == 0) return;

	//start progress reporter
	ProgressReporter progress(nCausticPhotons+nDirectPhotons+ // NOBOOK
		nIndirectPhotons, "Shooting photons"); // NOBOOK

	//create empty vectors of photons
	vector<Photon> causticPhotons;
	vector<Photon> directPhotons;
	vector<Photon> indirectPhotons;

	//reserve correct sizes
	causticPhotons.reserve(nCausticPhotons); // NOBOOK
	directPhotons.reserve(nDirectPhotons); // NOBOOK
	indirectPhotons.reserve(nIndirectPhotons); // NOBOOK

	// Initialize photon shooting statistics
	static StatsCounter nshot("Photon Map",
		"Number of photons shot from lights");

	//set up booleans for when we have enough photons for each type of ray
	bool causticDone = (nCausticPhotons == 0);
	bool directDone = (nDirectPhotons == 0);
	bool indirectDone = (nIndirectPhotons == 0);




#ifdef COMBINE_WITH_PHOTON_SURFACE

	for(int a=0;a<addedPhotons.size();a++)
	{
		Photon* photon = addedPhotons[a];
		Photon npPhoton(photon->p, photon->alpha, photon->wi);
		delete photon;
		indirectPhotons.push_back(npPhoton);
	}

	addedPhotons.clear();
	if (indirectPhotons.size() == nIndirectPhotons) {
		indirectDone = true;
		nIndirectPaths = nIndirectPathsSoFar;
		indirectMap =
			new KdTree<Photon,
					   PhotonProcess>(indirectPhotons);
	}

#endif


	
	//while we still have rays left
	while (!causticDone || !directDone || !indirectDone) {

		//we shot another ray
		++nshot;

		// Give up if we're not storing enough photons
		//shot 500000 but didn't save enough?
		if (nshot > 1500000 &&
			(unsuccessful(nCausticPhotons,
			              causticPhotons.size(),
						  nshot) ||
			 unsuccessful(nDirectPhotons,
			              directPhotons.size(),
						  nshot) ||
			 unsuccessful(nIndirectPhotons,
			              indirectPhotons.size(),
						  nshot))) {
			Error("Unable to store enough photons.  Giving up.\n");
			printf("Tried to shoot: %d photons, ended up with:\n\t%d (%d) caustic\n\t%d (%d) direct\n\t%d (%d) indirect\n",
				(int)nshot, causticPhotons.size(), nCausticPhotons, 
				directPhotons.size(), nDirectPhotons, indirectPhotons.size(), nIndirectPhotons);
			return;
		}

		// Trace a photon path and store contribution
		// Choose 4D sample values for photon

		//grab values based on the nshot value?
		float u[4];
		u[0] = (float)RadicalInverse((int)nshot+1, 2);
		u[1] = (float)RadicalInverse((int)nshot+1, 3);
		u[2] = (float)RadicalInverse((int)nshot+1, 5);
		u[3] = (float)RadicalInverse((int)nshot+1, 7);


		// Choose light to shoot photon from
		//get total number of lights
		int nLights = int(scene->lights.size());

		//pick a light
		int lightNum =
			min(Floor2Int(nLights * (float)RadicalInverse((int)nshot+1, 11)),
			nLights-1);

		if(ignoreFirstLight)
		{
			nLights--;
			lightNum =
				1+min(Floor2Int(nLights * (float)RadicalInverse((int)nshot+1, 11)),
				nLights-1);
		}

		//get a pointer to the light
		Light *light = scene->lights[lightNum];

		//the light pdf is 1/total lights (evenly weighted)
		float lightPdf = 1.f / nLights;


		// Generate _photonRay_ from light source and initialize _alpha_
		//grab the photon ray
		RayDifferential photonRay;
		
		//save off a pdf
		float pdf;

		//grab the alpha spectrum from the light ray
		Spectrum alpha =
			light->Sample_L(scene, u[0], u[1], u[2], u[3],
				&photonRay, &pdf);

		//if there is no chance this ray helps or there is no light continue around again
		if (pdf == 0.f || alpha.Black()) continue;

		//attenuate the alpha by the pdf values.
		alpha /= pdf * lightPdf;

		//if there is actually some alpha
		if (!alpha.Black()) {
			// Follow photon path through scene and record intersections

			//set it at no specular path
			bool specularPath = false;

			//set up intersection data
			Intersection photonIsect;
			int nIntersections = 0;

			//while we find intersections for photonRay...
			while (scene->Intersect(photonRay, &photonIsect)) {

				//add one to the intersection counter
				++nIntersections;

				// Handle photon/surface intersection

				//attenuate the alpha by the transmittance of the volume
				alpha *= scene->Transmittance(photonRay);

				//store the reverse vector of the direction into wo
				Vector wo = -photonRay.d;

				//get the bsdf from the geometry that was hit
				BSDF *photonBSDF = photonIsect.GetBSDF(photonRay);

				//set up a specular type?
				BxDFType specularType = BxDFType(BSDF_REFLECTION |
					BSDF_TRANSMISSION | BSDF_SPECULAR);

				//does the bsdf have components that are not specular
				bool hasNonSpecular = (photonBSDF->NumComponents() >
					photonBSDF->NumComponents(specularType));

				//if we have non specular components...
				if (hasNonSpecular) {
					// Deposit photon at surface

					//create a photon shooting from the intersection point
					Photon photon(photonIsect.dg.p, alpha, wo);

					//if this is the first intersection...
					if (nIntersections == 1) {
						// Process direct lighting photon intersection

						//if we're not done keeping track of direct photons...
						if (!directDone) {

							//add the photon
							directPhotons.push_back(photon);

							//if the direct photons is full
							if (directPhotons.size() == nDirectPhotons) {
								//set it to done and keep track of the direct paths?
								directDone = true;
								nDirectPaths = (int)nshot;
								//set up the kd tree for direct photons
								directMap =
									new KdTree<Photon,
											   PhotonProcess>(directPhotons);
							}
							progress.Update(); // NOBOOK
						}
					}	//basically are we on a bounce?
					else if (specularPath) {
						// Process caustic photon intersection

						//if we have more caustic photons to process...
						if (!causticDone) {

							//add the photon
							causticPhotons.push_back(photon);

							//if the caustic photons is full
							if (causticPhotons.size() == nCausticPhotons) {
								//set it to done and keep track of the caustic paths?
								causticDone = true;
								nCausticPaths = (int)nshot;
								//set up the kd tree for caustic photons
								causticMap =
									new KdTree<Photon,
										       PhotonProcess>(causticPhotons);
							}
							progress.Update();
						}
					}	//something else... indirect lighting?
					else {
						// Process indirect lighting photon intersection

						//we done indirect lighting?
						if (!indirectDone) {
							//etc etc...
							indirectPhotons.push_back(photon);
							if (indirectPhotons.size() == nIndirectPhotons) {
								indirectDone = true;
#ifdef COMBINE_WITH_PHOTON_SURFACE
								nIndirectPaths = (int)nshot + nIndirectPathsSoFar;
#else
								nIndirectPaths = (int)nshot;
#endif
								indirectMap =
									new KdTree<Photon,
											   PhotonProcess>(indirectPhotons);
							}
							progress.Update();
						}
					}
				}

				// Sample new photon ray direction

				//set up some more vectors
				Vector wi;
				float pdf;
				BxDFType flags;
				// Get random numbers for sampling outgoing photon direction
				float u1, u2, u3;
				if (nIntersections == 1) {
					u1 = (float)RadicalInverse((int)nshot+1, 13);
					u2 = (float)RadicalInverse((int)nshot+1, 17);
					u3 = (float)RadicalInverse((int)nshot+1, 19);
				}
				else {
					u1 = RandomFloat();
					u2 = RandomFloat();
					u3 = RandomFloat();
				}

				//sample the photon's bsdf from random numbers
				Spectrum fr = photonBSDF->Sample_f(wo, &wi, u1, u2, u3,
					&pdf, BSDF_ALL, &flags);

				//if the value is black or contributes nothing break???
				if (fr.Black() || pdf == 0.f)
					break;

				//set the specular path if we're about to go on a second bounce or it's already been set
					//and make sure we have a specular component
				specularPath = (nIntersections == 1 || specularPath) &&
					((flags & BSDF_SPECULAR) != 0);

				//update the alpha based on the reflected bsdf spectrum and compute the weight?
				alpha *= fr * AbsDot(wi, photonBSDF->dgShading.nn) / pdf;

				//calculate the new photon ray from the photon's bsdf sample
				photonRay = RayDifferential(photonIsect.dg.p, wi);

				// Possibly terminate photon path

				//if we've done a lot of intersections, have a probability of stopping
				if (nIntersections > 3) {
					float continueProbability = .5f;
					if (RandomFloat() > continueProbability)
						break;
					//if we didn't stop, multiply the alpha by 2!?!?
					alpha /= continueProbability;
				}
			}
		}
		BSDF::FreeAll();
	}
	progress.Done(); // NOBOOK
}
Spectrum PhotonIntegrator::Li(const Scene *scene,
		const RayDifferential &ray, const Sample *sample,
		float *alpha) const {
	// Compute reflected radiance with photon map

	//set up the spectrum and intersection
	Spectrum L(0.);
	Intersection isect;

	//if we have an intersection with the scene...
	if (scene->Intersect(ray, &isect)) {

		//if they want alpha, set it to 1 to start
		if (alpha) *alpha = 1.;

		//keep track of the reverse direction vector
		Vector wo = -ray.d;

		// Compute emitted light if ray hit an area light source

		//get the emitted light given off in the direction of the ray
		L += isect.Le(wo);

		// Evaluate BSDF at hit point

		//get the bsdf of the hit surface
		BSDF *bsdf = isect.GetBSDF(ray);

		//grab some surface info?
		const Point &p = bsdf->dgShading.p;
		const Normal &n = bsdf->dgShading.nn;

		// Compute direct lighting for photon map integrator

		//(use only photons or not for direct lighting)
		//if we're computing direct with photons (?)
		if (directWithPhotons)	//add the result of LPhoton to the luminance
			L += LPhoton(directMap, nDirectPaths, nLookup,
				bsdf, isect, wo, maxDistSquared);
		else	//otherwise add light a sampling value
			L += UniformSampleAllLights(scene, p, n,
				wo, bsdf, sample,
				lightSampleOffset, bsdfSampleOffset,
				bsdfComponentOffset);
		
		// Compute indirect lighting for photon map integrator

		//add in indirect lighting
		L += LPhoton(causticMap, nCausticPaths, nLookup, bsdf,
			isect, wo, maxDistSquared);

		//if this is the final gather
		if (finalGather) {
			// Do one-bounce final gather for photon map

			//set the Li to 0
			Spectrum Li(0.);

			//loop through on gather samples
			for (int i = 0; i < gatherSamples; ++i) {
				// Sample random direction for final gather ray

				//set up some random numbers
				Vector wi;
				float u1 = sample->twoD[gatherSampleOffset][2*i];
				float u2 = sample->twoD[gatherSampleOffset][2*i+1];
				float u3 = sample->oneD[gatherComponentOffset][i];
				float pdf;

				//grab a sample from the bsdf
				Spectrum fr = bsdf->Sample_f(wo, &wi, u1, u2, u3,
					&pdf, BxDFType(BSDF_ALL & (~BSDF_SPECULAR)));

				//if the spectrum was black or the pdf is 0 continue...
				if (fr.Black() || pdf == 0.f) continue;

				//set up a bounce ray...
				RayDifferential bounceRay(p, wi);

				//keep track of gather rays...
				static StatsCounter gatherRays("Photon Map", // NOBOOK
					"Final gather rays traced"); // NOBOOK
				++gatherRays; // NOBOOK

				//find an intersection for the bounced ray
				Intersection gatherIsect;
				if (scene->Intersect(bounceRay, &gatherIsect)) {
					// Compute exitant radiance at final gather intersection

					//the bounce'd ray's bsdf
					BSDF *gatherBSDF = gatherIsect.GetBSDF(bounceRay);

					//the bounce direction negated
					Vector bounceWo = -bounceRay.d;

					//calculate the indirect luminance (?)
					Spectrum Lindir =
						LPhoton(directMap, nDirectPaths, nLookup,
							gatherBSDF, gatherIsect, bounceWo, maxDistSquared) +
						LPhoton(indirectMap, nIndirectPaths, nLookup,
							gatherBSDF, gatherIsect, bounceWo, maxDistSquared) +
						LPhoton(causticMap, nCausticPaths, nLookup,
							gatherBSDF, gatherIsect, bounceWo, maxDistSquared);

					//multiply it by the transmittance...
					Lindir *= scene->Transmittance(bounceRay);

					//add it, weighted, into the final value
					Li += fr * Lindir * AbsDot(wi, n) / pdf;
				}
			}
			//add the Li value to the L value but weighted by the number of gathered samples
			L += Li / float(gatherSamples);
		}
		else	//not doing final gather, so just grab LPhoton on the indirect map
			L += LPhoton(indirectMap, nIndirectPaths, nLookup,
				bsdf, isect, wo, maxDistSquared);

		//if we're not at the max specular depth yet
		if (specularDepth++ < maxSpecularDepth) {
			Vector wi;
			// Trace rays for specular reflection and refraction

			//grab a sample based on the bsdf
			Spectrum f = bsdf->Sample_f(wo, &wi,
				BxDFType(BSDF_REFLECTION | BSDF_SPECULAR));

			//if the spectrum isn't black...
			if (!f.Black()) {
				// Compute ray differential _rd_ for specular reflection

				//find the ray differential
				RayDifferential rd(p, wi);
				rd.hasDifferentials = true;
				rd.rx.o = p + isect.dg.dpdx;
				rd.ry.o = p + isect.dg.dpdy;

				// Compute differential reflected directions

				//need to get the differential reflected directions 
				Normal dndx = bsdf->dgShading.dndu * bsdf->dgShading.dudx +
					bsdf->dgShading.dndv * bsdf->dgShading.dvdx;
				Normal dndy = bsdf->dgShading.dndu * bsdf->dgShading.dudy +
					bsdf->dgShading.dndv * bsdf->dgShading.dvdy;
				Vector dwodx = -ray.rx.d - wo, dwody = -ray.ry.d - wo;
				float dDNdx = Dot(dwodx, n) + Dot(wo, dndx);
				float dDNdy = Dot(dwody, n) + Dot(wo, dndy);
				rd.rx.d = wi -
				          dwodx + 2 * Vector(Dot(wo, n) * dndx +
						  dDNdx * n);
				rd.ry.d = wi -
				          dwody + 2 * Vector(Dot(wo, n) * dndy +
						  dDNdy * n);

				//grab the scene's li from the ray differential and the samples
				L += scene->Li(rd, sample) * f * AbsDot(wi, n);
			}

			//grab another sample
			f = bsdf->Sample_f(wo, &wi,
				BxDFType(BSDF_TRANSMISSION | BSDF_SPECULAR));

			//if it's not black...
			if (!f.Black()) {
				// Compute ray differential _rd_ for specular transmission
				RayDifferential rd(p, wi);
				rd.hasDifferentials = true;
				rd.rx.o = p + isect.dg.dpdx;
				rd.ry.o = p + isect.dg.dpdy;
				
				float eta = bsdf->eta;
				Vector w = -wo;
				if (Dot(wo, n) < 0) eta = 1.f / eta;
				
				Normal dndx = bsdf->dgShading.dndu * bsdf->dgShading.dudx + bsdf->dgShading.dndv * bsdf->dgShading.dvdx;
				Normal dndy = bsdf->dgShading.dndu * bsdf->dgShading.dudy + bsdf->dgShading.dndv * bsdf->dgShading.dvdy;
				
				Vector dwodx = -ray.rx.d - wo, dwody = -ray.ry.d - wo;
				float dDNdx = Dot(dwodx, n) + Dot(wo, dndx);
				float dDNdy = Dot(dwody, n) + Dot(wo, dndy);
				
				float mu = eta * Dot(w, n) - Dot(wi, n);
				float dmudx = (eta - (eta*eta*Dot(w,n))/Dot(wi, n)) * dDNdx;
				float dmudy = (eta - (eta*eta*Dot(w,n))/Dot(wi, n)) * dDNdy;
				
				rd.rx.d = wi + eta * dwodx - Vector(mu * dndx + dmudx * n);
				rd.ry.d = wi + eta * dwody - Vector(mu * dndy + dmudy * n);
				L += scene->Li(rd, sample) * f * AbsDot(wi, n);
			}
		}
		--specularDepth;
	}
	else {
		// Handle ray with no intersection
		if (alpha) *alpha = 0.;
		//set teh alpha to 0 unless the lights of the scene contribute something
		for (u_int i = 0; i < scene->lights.size(); ++i)
			L += scene->lights[i]->Le(ray);
		if (alpha && !L.Black()) *alpha = 1.;
		return L;
	}
	return L;
}
Spectrum PhotonIntegrator::LPhoton(
		KdTree<Photon, PhotonProcess> *map,
		int nPaths, int nLookup, BSDF *bsdf,
		const Intersection &isect, const Vector &wo,
		float maxDistSquared) {

	//set up the spectrum and make sure we have a kd tree...
	Spectrum L(0.);
	if (!map) return L;

	//set up a bsdf of non specular components
	BxDFType nonSpecular = BxDFType(BSDF_REFLECTION |
		BSDF_TRANSMISSION | BSDF_DIFFUSE | BSDF_GLOSSY);

	//if there are no non-specular components return L
	if (bsdf->NumComponents(nonSpecular) == 0)
		return L;

	//keep track of total lookups
	static StatsCounter lookups("Photon Map", "Total lookups"); // NOBOOK
	// Initialize _PhotonProcess_ object, _proc_, for photon map lookups

	//initialize the photon lookup process...
	PhotonProcess proc(nLookup, isect.dg.p);

	//set up space for the close photons
	proc.photons =
		(ClosePhoton *)alloca(nLookup * sizeof(ClosePhoton));
	// Do photon map lookup
	++lookups;  // NOBOOK

	//look up the photons based on the intersection point
	map->Lookup(isect.dg.p, proc, maxDistSquared);


	// Accumulate light from nearby photons

	//accumulate light
	static StatsRatio foundRate("Photon Map", "Photons found per lookup"); // NOBOOK
	foundRate.Add(proc.foundPhotons, 1); // NOBOOK

	//calculate the weight scale based on the paths and the distance squared
	float scale = 1.f / (float(nPaths) * maxDistSquared * M_PI);

	// Estimate reflected light from photons

	//grab a pointer to the close photons
	ClosePhoton *photons = proc.photons;

	//save off the number of found photons
	int nFound = proc.foundPhotons;

	//calculate nf from the bsdf's normal?
	Normal Nf = Dot(wo, bsdf->dgShading.nn) < 0 ? -bsdf->dgShading.nn :
		bsdf->dgShading.nn;

	//if the bsdf has components that are reflection, transmission, or glossy...
	if (bsdf->NumComponents(BxDFType(BSDF_REFLECTION |
			BSDF_TRANSMISSION | BSDF_GLOSSY)) > 0) {
		// Compute exitant radiance from photons for glossy surface

		//go through all the found photons
		for (int i = 0; i < nFound; ++i) {
			//set up a flag based on the dot product of the normal and the photon's incoming direction
				//reflection if it's positive, transmission if it's negative
			BxDFType flag = Dot(Nf, photons[i].photon->wi) > 0.f ?
				BSDF_ALL_REFLECTION : BSDF_ALL_TRANSMISSION;

			//get the bsdf's f sample and scale it based on the alpha and the flag stuff??
			L += bsdf->f(wo, photons[i].photon->wi,	flag) *
				(scale * photons[i].photon->alpha);
		}
	}	//diffuse only?
	else {
		// Compute exitant radiance from photons for diffuse surface

		//go through found photons
		Spectrum Lr(0.), Lt(0.);
		for (int i = 0; i < nFound; ++i)	//if the normal do thte photon direction are positive
			if (Dot(Nf, photons[i].photon->wi) > 0.f)
				Lr += photons[i].photon->alpha;		//add the alpha value of the photon
			else
				Lt += photons[i].photon->alpha;		//add the alpha value of the photon?

		//add in the values
		L += (scale * INV_PI) * (Lr * bsdf->rho(wo, BSDF_ALL_REFLECTION) +
			Lt * bsdf->rho(wo, BSDF_ALL_TRANSMISSION));
	}
	return L;
}
#ifndef COMBINE_WITH_PHOTON_SURFACE
PhotonProcess::PhotonProcess(u_int mp, const Point &P)
	: p(P) {
	photons = 0;
	nLookup = mp;
	foundPhotons = 0;
}

void PhotonProcess::operator()(const Photon &photon,
		float distSquared, float &maxDistSquared) const {
	static StatsPercentage discarded("Photon Map", "Discarded photons"); // NOBOOK
	discarded.Add(0, 1); // NOBOOK
	if (foundPhotons < nLookup) {
		// Add photon to unordered array of photons
		photons[foundPhotons++] = ClosePhoton(&photon, distSquared);
		if (foundPhotons == nLookup) {
			std::make_heap(&photons[0], &photons[nLookup]);
			maxDistSquared = photons[0].distanceSquared;
		}
	}
	else {
		// Remove most distant photon from heap and add new photon
		discarded.Add(1, 0); // NOBOOK
		std::pop_heap(&photons[0], &photons[nLookup]);
		photons[nLookup-1] = ClosePhoton(&photon, distSquared);
		std::push_heap(&photons[0], &photons[nLookup]);
		maxDistSquared = photons[0].distanceSquared;
	}
}
#endif
extern "C" DLLEXPORT SurfaceIntegrator *CreateSurfaceIntegrator(const ParamSet &params) {
	int nCaustic = params.FindOneInt("causticphotons", 20000);
	int nDirect = params.FindOneInt("directphotons", 100000);
	int nIndirect = params.FindOneInt("indirectphotons", 100000);
	int nUsed = params.FindOneInt("nused", 50);
	int maxDepth = params.FindOneInt("maxdepth", 5);
	bool finalGather = params.FindOneBool("finalgather", true);
	bool directPhotons = params.FindOneBool("directwithphotons", false);
	int gatherSamples = params.FindOneInt("finalgathersamples", 32);
	float maxDist = params.FindOneFloat("maxdist", .1f);
	bool ignoreFirst = params.FindOneBool("ignorefirstlight", false);
	return new PhotonIntegrator(nCaustic, nDirect, nIndirect,
		nUsed, maxDepth, maxDist, finalGather, gatherSamples,
		directPhotons, ignoreFirst);
}
