
/*
 * pbrt source code Copyright(c) 1998-2005 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For educational use only; commercial use expressly forbidden.
 * NO WARRANTY, express or implied, for this software.
 * (See file License.txt for complete license)
 */

// single.cpp*
#include "volume.h"
#include "transport.h"
#include "scene.h"
// SingleScattering Declarations
class SingleScattering : public VolumeIntegrator {
public:
	// SingleScattering Public Methods
	SingleScattering(float ss) { stepSize = ss; }
	Spectrum Transmittance(const Scene *, const Ray &ray,
		const Sample *sample, float *alpha) const;
	void RequestSamples(Sample *sample, const Scene *scene);
	Spectrum Li(const Scene *, const RayDifferential &ray, const Sample *sample, float *alpha) const;
private:
	// SingleScattering Private Data
	float stepSize;
	int tauSampleOffset, scatterSampleOffset;
};
// SingleScattering Method Definitions
void SingleScattering::RequestSamples(Sample *sample,
		const Scene *scene) {
	tauSampleOffset = sample->Add1D(1);
	scatterSampleOffset = sample->Add1D(1);
}

Spectrum SingleScattering::Transmittance(const Scene *scene,
		const Ray &ray, const Sample *sample, float *alpha) const {

	//if there is no volume region return 0
	if (!scene->volumeRegion) return Spectrum(1.f);

	//calculate the step based on if we have a sample pointer
	float step = sample ? stepSize : 4.f * stepSize;

	//calculate the offset from the sample or from a random float
	float offset = sample ? sample->oneD[tauSampleOffset][0] :
		RandomFloat();

	//calculate tau (see equations)
	Spectrum tau = scene->volumeRegion->Tau(ray, step, offset);

	//return e^-tau
	return Exp(-tau);
}

Spectrum SingleScattering::Li(const Scene *scene,
		const RayDifferential &ray, const Sample *sample,
		float *alpha) const {
	//grab the volume region from the scene
	VolumeRegion *vr = scene->volumeRegion;

	//grab the start and end times from the ray where it intersects the volume region
	float t0, t1;

	//if there is no volume region or the ray does not intersect it, return 0
	if (!vr || !vr->IntersectP(ray, &t0, &t1)) return 0.f;

	// Do single scattering volume integration in _vr_
	Spectrum Lv(0.);
	// Prepare for volume integration stepping

	//figure out how many steps to do based on the 
	int N = Ceil2Int((t1-t0) / stepSize);

	//get how far each particular step is
	float step = (t1 - t0) / N;

	//set up T, the starting poitn p, the prevoius point, and the direction back to the ray's origin
	Spectrum Tr(1.f);
	Point p = ray(t0), pPrev;
	Vector w = -ray.d;

	//if we have a sample grab a random offset from the sampler
	if (sample)
		t0 += sample->oneD[scatterSampleOffset][0] * step;
	else
		t0 += RandomFloat() * step;

	// Compute sample patterns for single scattering samples

	//allocate space for samples
	float *samp = (float *)alloca(3 * N * sizeof(float));

	//what the hell is this?
	LatinHypercube(samp, N, 3);

	//set the sample offset to 0
	int sampOffset = 0;

	//loop through all the steps
	for (int i = 0; i < N; ++i, t0 += step) {

		// Advance to sample at _t0_ and update _T_

		//save off the previous point
		pPrev = p;

		//get the next point
		p = ray(t0);

		//grab the tau of the step
		Spectrum stepTau = vr->Tau(Ray(pPrev, p - pPrev, 0, 1),
			.5f * stepSize, RandomFloat());

		//multiply that into Tr?
		Tr *= Exp(-stepTau);

		//if transmittance is really small, have a chance of terminating, otherwise multiply Tr by 2
		// Possibly terminate raymarching if transmittance is small
		if (Tr.y() < 1e-3) {
			const float continueProb = .5f;
			if (RandomFloat() > continueProb) break;
			Tr /= continueProb;
		}
		// Compute single-scattering source term at _p_

		//grab the emitted light with respect to the density at the point in the direction of the eye
		Lv += Tr * vr->Lve(p, w);

		//grab the scatter spetrum at the point
		Spectrum ss = vr->sigma_s(p, w);

		//if we have lights and the scattered spectrum is not black
		if (!ss.Black() && scene->lights.size() > 0) {
			//grab the number of lights
			int nLights = scene->lights.size();

			//grab a particular light based on samples
			int lightNum =
				min(Floor2Int(samp[sampOffset] * nLights),
				    nLights-1);
			Light *light = scene->lights[lightNum];

			// Add contribution of _light_ due to scattering at _p_

			//figure out the light's contribution
			float pdf;
			VisibilityTester vis;
			Vector wo;
			//grab two random numbers
			float u1 = samp[sampOffset+1], u2 = samp[sampOffset+2];

			//sample the light given the random numbers
			Spectrum L = light->Sample_L(p, u1, u2, &wo, &pdf, &vis);

			//if the light gives off light and is visible
			if (!L.Black() && pdf > 0.f && vis.Unoccluded(scene)) {
				//mulitply by transmittance so that we accurately update the light value
				Spectrum Ld = L * vis.Transmittance(scene);
				//calculate the updated Lv
				Lv += Tr * ss * vr->p(p, w, -wo) *
				      Ld * float(nLights) / pdf;
			}
		}
		//move to the next set of samples
		sampOffset += 3;
	}
	//multiply by step
	return Lv * step;
}
extern "C" DLLEXPORT VolumeIntegrator *CreateVolumeIntegrator(const ParamSet &params) {
	float stepSize  = params.FindOneFloat("stepsize", 1.f);
	return new SingleScattering(stepSize);
}
