
/*
 * pbrt source code Copyright(c) 1998-2005 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For educational use only; commercial use expressly forbidden.
 * NO WARRANTY, express or implied, for this software.
 * (See file License.txt for complete license)
 */

//#define USE_ADAPTIVE_RAY_MARCHING_FOR_INTERACTION

// single.cpp*
#include "volume.h"
#include "pbrt.h"
#include "transport.h"
#include "scene.h"
#include "mc.h"
#include "kdtree.h"
#include "sampling.h"

#ifdef COMBINE_WITH_PHOTON_SURFACE
#include "photonmap.h"
#endif

#include <fstream>
#include <time.h>

// SingleScattering Declarations
#ifndef COMBINE_WITH_PHOTON_SURFACE
struct Photon;
struct PhotonProcess;
struct ClosePhoton;
#endif

class VolumePhotonIntegrator : public VolumeIntegrator {
public:
	// SingleScattering Public Methods
	VolumePhotonIntegrator(float ss, float maxDist, bool directPhotons, int count, float ig, int lookup, bool ignoreFirst);
	~VolumePhotonIntegrator();
	Spectrum Transmittance(const Scene *, const Ray &ray,
		const Sample *sample, float *alpha) const;
	void RequestSamples(Sample *sample, const Scene *scene);
	Spectrum Li(const Scene *, const RayDifferential &ray, const Sample *sample, float *alpha) const;

	void Preprocess(const Scene *);

	Spectrum LPhotonV(KdTree<Photon, PhotonProcess> *map,
		int nPaths, int nLookup, VolumeRegion* vr, const Point &p,
		const Vector &w, float maxDistCubed, float maxDistSquared) const;

	static inline bool unsuccessful(int needed, int found, int shot) {
		return (found < needed &&
		        (found == 0 || found < shot / 1024));
	}
private:
	void WritePhotonsOutToMap(std::vector<Photon>* map);
	// SingleScattering Private Data
	float stepSize;
	int tauSampleOffset, scatterSampleOffset;
	
	bool directWithPhotons;
	mutable KdTree<Photon, PhotonProcess> *photonMap;
	u_int nLookup;
	int nPhotonPaths;
	u_int nPhotons;
	float maxDistSquared;
	float maxDistCubed;
	float g;
	bool ignoreFirstLight;
};

#ifndef COMBINE_WITH_PHOTON_SURFACE
struct Photon {
	// Photon Constructor
	Photon(const Point &pp, const Spectrum &wt, const Vector &w)
		: p(pp), alpha(wt), wi(w) {
	}

	Photon() { }

	Point p;
	Spectrum alpha;
	Vector wi;
};

struct PhotonProcess {
	// PhotonProcess Public Methods
	PhotonProcess(u_int mp, const Point &p);

	void operator()(const Photon &photon, float dist2, float &maxDistSquared) const;

	const Point &p;
	ClosePhoton *photons;
	u_int nLookup;
	mutable u_int foundPhotons;
};

struct ClosePhoton {
	ClosePhoton(const Photon *p = NULL,
	            float md2 = INFINITY) {
		photon = p;
		distanceSquared = md2;
	}

	bool operator<(const ClosePhoton &p2) const {
		return distanceSquared == p2.distanceSquared ? (photon < p2.photon) :
			distanceSquared < p2.distanceSquared;
	}

	const Photon *photon;
	float distanceSquared;
};

PhotonProcess::PhotonProcess(u_int mp, const Point &P)
	: p(P) {
	photons = 0;
	nLookup = mp;
	foundPhotons = 0;
}

void PhotonProcess::operator()(const Photon &photon,
		float distSquared, float &maxDistSquared) const {
	static StatsPercentage discarded("Volumetric Photon Map", "Discarded photons"); // NOBOOK
	discarded.Add(0, 1); // NOBOOK
	if (foundPhotons < nLookup) {
		// Add photon to unordered array of photons
		photons[foundPhotons++] = ClosePhoton(&photon, distSquared);
		if (foundPhotons == nLookup) {
			std::make_heap(&photons[0], &photons[nLookup]);
			maxDistSquared = photons[0].distanceSquared;
		}
	}
	else {
		// Remove most distant photon from heap and add new photon
		discarded.Add(1, 0); // NOBOOK
		std::pop_heap(&photons[0], &photons[nLookup]);
		photons[nLookup-1] = ClosePhoton(&photon, distSquared);
		std::push_heap(&photons[0], &photons[nLookup]);
		maxDistSquared = photons[0].distanceSquared;
	}
}
#endif

VolumePhotonIntegrator::VolumePhotonIntegrator(float ss, float maxDist, bool directPhotons, int count, float ig, int lookup, bool ignoreFirst)
{
	stepSize = ss;
	maxDistSquared = maxDist*maxDist;
	maxDistCubed = maxDist*maxDist*maxDist;
	directWithPhotons = directPhotons;
	nPhotons = count;
	g = ig;
	nLookup = lookup;
	ignoreFirstLight = ignoreFirst;
}

// SingleScattering Method Definitions
void VolumePhotonIntegrator::RequestSamples(Sample *sample,
		const Scene *scene) {
	tauSampleOffset = sample->Add1D(1);
	scatterSampleOffset = sample->Add1D(1);
	photonMap = NULL;
	directWithPhotons = false;
}

VolumePhotonIntegrator::~VolumePhotonIntegrator()
{
	delete photonMap;
}

void VolumePhotonIntegrator::Preprocess(const Scene *scene)
{
	//shoot photons into the volume...
	//take directWithPhotons into account
#ifdef COMBINE_WITH_PHOTON_SURFACE
	int indirectSurfacePhotonCount = 0;
#endif

	VolumeRegion *vr = scene->volumeRegion;
	if(!vr)
		return;

	if (scene->lights.size() == 0) return;
	
	//start progress reporter
	ProgressReporter progress(nPhotons, "Shooting vol. photons"); // NOBOOK
	
	vector<Photon> photons;
	photons.reserve(nPhotons);
	static StatsCounter nshot("Volumetric Photon Map",
		"Number of photons shot from lights");
	bool photonsDone = (nPhotons == 0);

	static StatsCounter surfacereflections("Volumetric Photon Map",
		"Photons reflected off surface");

	while(!photonsDone)
	{
		++nshot;
		if (nshot > 1500000 &&
			unsuccessful(nPhotons,
			              photons.size(),
						  nshot)) {
			Error("Unable to store enough photons.  Giving up.\n");
			return;
		}
		float u[4];
		u[0] = (float)RadicalInverse((int)nshot+1, 2);
		u[1] = (float)RadicalInverse((int)nshot+1, 3);
		u[2] = (float)RadicalInverse((int)nshot+1, 5);
		u[3] = (float)RadicalInverse((int)nshot+1, 7);

		int nLights = int(scene->lights.size());
		int lightNum =
			min(Floor2Int(nLights * (float)RadicalInverse((int)nshot+1, 11)),
			nLights-1);

		if(ignoreFirstLight)
		{
			nLights--;
			lightNum = 1+min((float)RadicalInverse((int)nshot+1, 11) * nLights, nLights-1);
		}


		Light *light = scene->lights[lightNum];
		float lightPdf = 1.f / nLights;
		RayDifferential photonRay;
		float pdf;
		Spectrum alpha = light->Sample_L(scene, u[0], u[1], u[2], u[3], &photonRay, &pdf);
		if (pdf == 0.f || alpha.Black()) continue;

		alpha /= pdf * lightPdf;
		//alpha *= pdf;

		if (!alpha.Black()) {
			Intersection photonIsect;
			int nIntersections = 0;
			//got a sample... now what?

			//loop 5 times for a volume intersection
				//do we intersect the volume?
				//do we intersect any scene objects?

				//if we hit the volume and not the scene, go to the main loop
				//if we hit the scene and not the volume, calculate a new ray
				//if we hit both, test to see if we hit the volume going back from the scene object
					//if we do hit the volume, we are inside, we can go to the next part
					//if not, calculate a new ray
			/*
			float t0, t1;
			bool hitNothing = false;
			for(int i=0;i<5;i++)
			{
				Vector wd = photonRay.d;
				bool hitVolume = false;
				bool hitScene = false;
				bool createNewRay = false;
				Intersection sceneInt;
				float bounce_pdf;
				if(vr->IntersectP(photonRay, &t0, &t1))
					hitVolume = true;
				if(t0 <= photonRay.mint + 0.0001f)
				{
					printf("already inside\n");
					break;
				}
				if(scene->Intersect(photonRay, &sceneInt))
					hitScene = true;
				
				if(hitVolume && !hitScene)
				{
					//we hit the volume and not the scene we're done
					if(t0 > 0)
						photonRay = RayDifferential(photonRay(t0), wd);
					break;
				} else if(!hitVolume && !hitScene)
				{
					//we hit nothing, we're done
					hitNothing = true;
					break;
				} else if(hitVolume && hitScene)
				{
					//we hit both, sooo lets see if we still have to do this again
					RayDifferential tmpRay = RayDifferential(sceneInt.dg.p, -photonRay.d);
					float t0_2, t1_2;
					if(vr->IntersectP(tmpRay, &t0_2, &t1_2))
					{
						//if(t0_2 <= tmpRay.mint + 0.0001f)
						//{
							//printf("volume first\n");
							if(t0 > 0)
								RayDifferential(photonRay(t0), wd);
							break;
						//}
					}
				}

				if(i == 4)
				{
					hitNothing = true;
					break;
				}
				//create a new ray
				float u1, u2, u3;
				if (i == 1) {
					u1 = (float)RadicalInverse((int)nshot+1, 13);
					u2 = (float)RadicalInverse((int)nshot+1, 17);
					u3 = (float)RadicalInverse((int)nshot+1, 19);
				} else {
					u1 = RandomFloat();
					u2 = RandomFloat();
					u3 = RandomFloat();
				}
				Vector wi;
				Vector wo = -photonRay.d;
				BxDFType flags;
				BSDF *surfaceBSDF = sceneInt.GetBSDF(photonRay);

				Spectrum fr = surfaceBSDF->Sample_f(wo, &wi, u1, u2, u3, &bounce_pdf, BSDF_ALL, &flags);

				if (fr.Black() || bounce_pdf == 0.f)
				{
					hitNothing = true;
					break;
				}
				alpha *= fr * AbsDot(wi, surfaceBSDF->dgShading.nn) / bounce_pdf;
				photonRay = RayDifferential(sceneInt.dg.p, wi);

			}
			if(hitNothing) continue;
			*/

			//other code
			float t0, t1;
			if(!vr->IntersectP(photonRay, &t0, &t1)) continue;

			//do we hit the scene before the fog?  If so stop...
			Intersection testIntersect;
			if(scene->Intersect(photonRay, &testIntersect))
			{
				float tmp_dist = sqrt((photonRay.o.x-testIntersect.dg.p.x)*(photonRay.o.x-testIntersect.dg.p.x)
											+(photonRay.o.y-testIntersect.dg.p.y)*(photonRay.o.y-testIntersect.dg.p.y)
											+(photonRay.o.z-testIntersect.dg.p.z)*(photonRay.o.z-testIntersect.dg.p.z));
				if(tmp_dist < t0)
				{
					continue;
				}
			}


			if(t0 > photonRay.mint + 0.0001f)
			{
				photonRay = RayDifferential(photonRay(t0), photonRay.d);
			}


			//main loop
				//do we intersect the volume (have we left it?)
				//calculate distance to next interaction
				//do we hit a surface before that?
					//if yes
						//bounce off surface, continue
				//use ratio to decide whether or not to scatter or absorb ray.  stick it in the volume either way
					//yes we scatter
						//redirect using p(), and continue
					//no we do not scatter it was absorbed
						//break

			int ctr = 0;
			int surface_bounce = 0;
			Ray tmpRay;
			float tti = 0.0f;
			while(vr->IntersectP(photonRay, &t0, &t1))
			{
				ctr += 1;
				//hmm keep this?
				if(ctr > 100)
				{
					//printf("Breaking because of counter\n");
					break;
				}
				//printf("Current counter: %d (with %d photons)\n", ctr, photons.size()); fflush(stdout);
#ifdef USE_ADAPTIVE_RAY_MARCHING_FOR_INTERACTION
				int N = Ceil2Int((t1-t0) / stepSize);
				float step = (t1 - t0) / N;
				t0 = stepSize;
				Point p(t0);
				float totalSigmaT = 0;
				float rand = RandomFloat();
				for (int i = 0; i < N; ++i, t0 += step) {
					float avg_sigma;
					p = photonRay(t0);
					totalSigmaT += vr->sigma_t(p, photonRay.d).y();
					avg_sigma = totalSigmaT/float(i+1);
					tti = -1 * log(rand)/avg_sigma;
					if(tti > t0)
						break;
				}
				if(N == 0)
					continue;
				
				//tti = -1 * log(rand)/vr->sigma_t(photonRay.o, photonRay.d).y();
#else
				//need to calculate a proper time to next interaction here...
				if(vr->sigma_t(photonRay.o, photonRay.d).y() > 0)
				{
					float rand = RandomFloat();
					tti = -1 * log(rand)/vr->sigma_t(photonRay.o, photonRay.d).y();
				} else {
					//ray march until it's not 0?
					photonRay = RayDifferential(photonRay(stepSize), photonRay.d);
					continue;
				}
#endif
				tmpRay.o = photonRay.o;
				tmpRay.d = photonRay.d;
				tmpRay.mint = RAY_EPSILON;
				tmpRay.maxt = tti;
				tmpRay.time = tmpRay.mint;
				
				float dist = 0;
				bool foundIntersect = false;
				Intersection photonIntersect;
				if(scene->Intersect(tmpRay, &photonIntersect))
				{
					dist = sqrt((photonRay.o.x-photonIntersect.dg.p.x)*(photonRay.o.x-photonIntersect.dg.p.x)
											+(photonRay.o.y-photonIntersect.dg.p.y)*(photonRay.o.y-photonIntersect.dg.p.y)
											+(photonRay.o.z-photonIntersect.dg.p.z)*(photonRay.o.z-photonIntersect.dg.p.z));
					foundIntersect = true;
				}

				//step and update the average sigma_t until -log(some randomFloat)/average-sigma_t <= that dist then check intersection dist


				
				//printf("Comparing %f > %f\n", tti, tmpRay.mint); fflush(stdout);
				if(tti > tmpRay.mint)
				{
					if(scene->Intersect(tmpRay, &photonIntersect))
					{
						float dist = sqrt((photonRay.o.x-photonIntersect.dg.p.x)*(photonRay.o.x-photonIntersect.dg.p.x)
											+(photonRay.o.y-photonIntersect.dg.p.y)*(photonRay.o.y-photonIntersect.dg.p.y)
											+(photonRay.o.z-photonIntersect.dg.p.z)*(photonRay.o.z-photonIntersect.dg.p.z));
						if(dist < tti)
						{
							//printf("creating a new ray due to %f < %f\n", dist, time_to_interact);
							surfacereflections++;
							float u1, u2, u3;
							if (ctr == 1) {
								u1 = (float)RadicalInverse((int)nshot+1, 13);
								u2 = (float)RadicalInverse((int)nshot+1, 17);
								u3 = (float)RadicalInverse((int)nshot+1, 19);
							} else {
								u1 = RandomFloat();
								u2 = RandomFloat();
								u3 = RandomFloat();
							}
							Vector wi;
							Vector wo = -photonRay.d;
							BxDFType flags;
							BSDF *surfaceBSDF = photonIntersect.GetBSDF(photonRay);
							float bounce_pdf;
							Spectrum fr = surfaceBSDF->Sample_f(wo, &wi, u1, u2, u3, &bounce_pdf, BSDF_ALL, &flags);
							if (fr.Black() || bounce_pdf == 0.f)
							{
								break;
							}
							alpha *= fr * AbsDot(wi, surfaceBSDF->dgShading.nn) / bounce_pdf;
							if(alpha.Black() || alpha < 0.1f)
							{
								break;
							}
							photonRay = RayDifferential(photonIntersect.dg.p, wi);
							surface_bounce++;

#ifdef COMBINE_WITH_PHOTON_SURFACE
							Photon* photonPtr = new Photon(photonIntersect.dg.p, alpha, wo);
							((PhotonIntegrator*)scene->surfaceIntegrator)->AddIndirectPhoton(photonPtr);
							indirectSurfacePhotonCount++;
#endif
							continue;
						}
					} else {
					}
				}
				



				photonRay = RayDifferential(photonRay(tti+photonRay.time), photonRay.d);

				//add ray!
				if(directWithPhotons || ctr > 1)
				{
					Photon photon(photonRay.o, alpha, -photonRay.d);
					photons.push_back(photon);
					progress.Update();
					if(photons.size() == nPhotons)
					{
						photonsDone = true;
						nPhotonPaths = (int)nshot;
						photonMap = new KdTree<Photon, PhotonProcess>(photons);
						WritePhotonsOutToMap(&photons);
						break;
					}
				}
				if(RandomFloat() <= vr->sigma_s(photonRay.o, photonRay.d).y()/vr->sigma_t(photonRay.o, photonRay.d).y())
				{
					float rand_float = RandomFloat();
					float cos_theta = (1/(2*g))*(1 + g*g - ((1-g*g)/(1-g+2*g*rand_float))*((1-g*g)/(1-g+2*g*rand_float)));
					if(cos_theta > 1)
						cos_theta = 1.f;
					float sin_theta = sqrt(1-cos_theta*cos_theta);

					float phi = RandomFloat();
					float sin_phi = sin(2*M_PI*phi);
					float cos_phi = cos(2*M_PI*phi);

					/*
					//theta around z
					//phi around y
					//godda multiply 2 matrices :(
					Matrix4x4 mat_phi = Matrix4x4(cos_phi, sin_phi, 0, 0,
												-sin_phi, cos_phi, 0, 0,
												0, 0, 1, 0,
												0, 0, 0, 1);
					Matrix4x4 mat_theta = Matrix4x4(cos_theta, 0, -sin_theta, 0,
												0, 1, 0, 0,
												sin_theta, 0, cos_theta, 0,
												0, 0, 0, 1);
					
					//phi * theta
					Matrix4x4 final_rot = Matrix4x4(mat_phi.m[0][0]*mat_theta.m[0][0] + mat_phi.m[0][1]*mat_theta.m[1][0] + mat_phi.m[0][2]*mat_theta.m[2][0] + mat_phi.m[0][3]*mat_theta.m[3][0],
													mat_phi.m[0][0]*mat_theta.m[0][1] + mat_phi.m[0][1]*mat_theta.m[1][1] + mat_phi.m[0][2]*mat_theta.m[2][1] + mat_phi.m[0][3]*mat_theta.m[3][1],
													mat_phi.m[0][0]*mat_theta.m[0][2] + mat_phi.m[0][1]*mat_theta.m[1][2] + mat_phi.m[0][2]*mat_theta.m[2][2] + mat_phi.m[0][3]*mat_theta.m[3][2],
													mat_phi.m[0][0]*mat_theta.m[0][3] + mat_phi.m[0][1]*mat_theta.m[1][3] + mat_phi.m[0][2]*mat_theta.m[2][3] + mat_phi.m[0][3]*mat_theta.m[3][3],
													mat_phi.m[1][0]*mat_theta.m[0][0] + mat_phi.m[1][1]*mat_theta.m[1][0] + mat_phi.m[1][2]*mat_theta.m[2][0] + mat_phi.m[1][3]*mat_theta.m[3][0],
													mat_phi.m[1][0]*mat_theta.m[0][1] + mat_phi.m[1][1]*mat_theta.m[1][1] + mat_phi.m[1][2]*mat_theta.m[2][1] + mat_phi.m[1][3]*mat_theta.m[3][1],
													mat_phi.m[1][0]*mat_theta.m[0][2] + mat_phi.m[1][1]*mat_theta.m[1][2] + mat_phi.m[1][2]*mat_theta.m[2][2] + mat_phi.m[1][3]*mat_theta.m[3][2],
													mat_phi.m[1][0]*mat_theta.m[0][3] + mat_phi.m[1][1]*mat_theta.m[1][3] + mat_phi.m[1][2]*mat_theta.m[2][3] + mat_phi.m[1][3]*mat_theta.m[3][3],
													mat_phi.m[2][0]*mat_theta.m[0][0] + mat_phi.m[2][1]*mat_theta.m[1][0] + mat_phi.m[2][2]*mat_theta.m[2][0] + mat_phi.m[2][3]*mat_theta.m[3][0],
													mat_phi.m[2][0]*mat_theta.m[0][1] + mat_phi.m[2][1]*mat_theta.m[1][1] + mat_phi.m[2][2]*mat_theta.m[2][1] + mat_phi.m[2][3]*mat_theta.m[3][1],
													mat_phi.m[2][0]*mat_theta.m[0][2] + mat_phi.m[2][1]*mat_theta.m[1][2] + mat_phi.m[2][2]*mat_theta.m[2][2] + mat_phi.m[2][3]*mat_theta.m[3][2],
													mat_phi.m[2][0]*mat_theta.m[0][3] + mat_phi.m[2][1]*mat_theta.m[1][3] + mat_phi.m[2][2]*mat_theta.m[2][3] + mat_phi.m[2][3]*mat_theta.m[3][3],
													mat_phi.m[3][0]*mat_theta.m[0][0] + mat_phi.m[3][1]*mat_theta.m[1][0] + mat_phi.m[3][2]*mat_theta.m[2][0] + mat_phi.m[3][3]*mat_theta.m[3][0],
													mat_phi.m[3][0]*mat_theta.m[0][1] + mat_phi.m[3][1]*mat_theta.m[1][1] + mat_phi.m[3][2]*mat_theta.m[2][1] + mat_phi.m[3][3]*mat_theta.m[3][1],
													mat_phi.m[3][0]*mat_theta.m[0][2] + mat_phi.m[3][1]*mat_theta.m[1][2] + mat_phi.m[3][2]*mat_theta.m[2][2] + mat_phi.m[3][3]*mat_theta.m[3][2],
													mat_phi.m[3][0]*mat_theta.m[0][3] + mat_phi.m[3][1]*mat_theta.m[1][3] + mat_phi.m[3][2]*mat_theta.m[2][3] + mat_phi.m[3][3]*mat_theta.m[3][3]);
					Vector d_old = photonRay.d;
					photonRay.d.x = final_rot.m[0][0]*d_old.x + final_rot.m[0][1]*d_old.y + final_rot.m[0][2]*d_old.z;
					photonRay.d.y = final_rot.m[1][0]*d_old.x + final_rot.m[1][1]*d_old.y + final_rot.m[1][2]*d_old.z;
					photonRay.d.z = final_rot.m[2][0]*d_old.x + final_rot.m[2][1]*d_old.y + final_rot.m[2][2]*d_old.z;
					*/

					//old
					float cos_theta_old = sqrt(photonRay.d.x*photonRay.d.x + photonRay.d.y*photonRay.d.y);
					float new_cos_theta = cos_theta + cos_theta_old;

					photonRay.d.z = sqrt(1-cos_theta*cos_theta);
					photonRay.d.x = cos_phi*new_cos_theta;
					photonRay.d.y = sin_phi*new_cos_theta;

				} else {
					break;
				}
			}
			//printf("Broke at %d with %d surface bounces\n", ctr, surface_bounce);
		}
	}
#ifdef COMBINE_WITH_PHOTON_SURFACE
	((PhotonIntegrator*)scene->surfaceIntegrator)->SetIndirectPathsSoFar(indirectSurfacePhotonCount);

#endif
	progress.Done();
}

Spectrum VolumePhotonIntegrator::Transmittance(const Scene *scene,
		const Ray &ray, const Sample *sample, float *alpha) const {

	if (!scene->volumeRegion) return Spectrum(1.f);
	float step = sample ? stepSize : 4.f * stepSize;
	float offset = sample ? sample->oneD[tauSampleOffset][0] :
		RandomFloat();
	Spectrum tau = scene->volumeRegion->Tau(ray, step, offset);
	return Exp(-tau);
}

Spectrum VolumePhotonIntegrator::Li(const Scene *scene,
		const RayDifferential &ray, const Sample *sample,
		float *alpha) const {
	//grab the volume region from the scene
	VolumeRegion *vr = scene->volumeRegion;

	float t0, t1;
	if (!vr || !vr->IntersectP(ray, &t0, &t1)) return 0.f;

	// Do single scattering volume integration in _vr_
	Spectrum Lv(0.);
	int N = Ceil2Int((t1-t0) / stepSize);
	float step = (t1 - t0) / N;
	Spectrum Tr(1.f);
	Point p = ray(t0), pPrev;
	Vector w = -ray.d;
	if (sample)
		t0 += sample->oneD[scatterSampleOffset][0] * step;
	else
		t0 += RandomFloat() * step;

	// Compute sample patterns for single scattering samples
	float *samp = (float *)alloca(3 * N * sizeof(float));
	LatinHypercube(samp, N, 3);
	int sampOffset = 0;
	for (int i = 0; i < N; ++i, t0 += step) {

		// Advance to sample at _t0_ and update _T_
		pPrev = p;
		p = ray(t0);
		Spectrum stepTau = vr->Tau(Ray(pPrev, p - pPrev, 0, 1),
			.5f * stepSize, RandomFloat());
		Tr *= Exp(-stepTau);
		if (Tr.y() < 1e-3) {
			const float continueProb = .5f;
			if (RandomFloat() > continueProb) break;
			Tr /= continueProb;
		}
		// Compute single-scattering source term at _p_
		Lv += Tr * vr->Lve(p, w);

		if(!directWithPhotons)
		{
			Spectrum ss = vr->sigma_s(p, w);
			if (!ss.Black() && scene->lights.size() > 0) {
				int nLights = scene->lights.size();
				int lightNum =
					min(Floor2Int(samp[sampOffset] * nLights),
						nLights-1);

				//nLights--;
				//lightNum = 1+min(Floor2Int(samp[sampOffset]* nLights), nLights-1);

				Light *light = scene->lights[lightNum];
				if(lightNum == 0)
				{
					static StatsCounter light0("Volumetric Photon Map",
						"Hit Light 0");
					++light0;
				} else if(lightNum == 1)
				{
					static StatsCounter light1("Volumetric Photon Map",
						"Hit Light 1");
					++light1;
				}
				// Add contribution of _light_ due to scattering at _p_
				float pdf;
				VisibilityTester vis;
				Vector wo;
				float u1 = samp[sampOffset+1], u2 = samp[sampOffset+2];
				Spectrum L = light->Sample_L(p, u1, u2, &wo, &pdf, &vis);
				if (!L.Black() && pdf > 0.f && vis.Unoccluded(scene)) {
					Spectrum Ld = L * vis.Transmittance(scene);
					Lv += Tr * ss * vr->p(p, w, -wo) *
						  Ld * float(nLights) / pdf;
				}
			}
		}

		//add photon mapping for scattering after the first scatter
		Lv += LPhotonV(photonMap, nPhotonPaths, nLookup,
			vr, p, w, maxDistCubed, maxDistSquared);


		//fix this if you use other sample offsets...
		sampOffset += 3;
	}
	return Lv * step;
}


Spectrum VolumePhotonIntegrator::LPhotonV(
		KdTree<Photon, PhotonProcess> *map,
		int nPaths, int nLookup, VolumeRegion* vr,
		const Point &p, const Vector &wo,
		float maxDistCubed, float maxDistSquared) const {

	Spectrum L(0.);
	if (!map) return L;

	static StatsCounter lookups("Volumetric Photon Map", "Total lookups"); // NOBOOK
	// Initialize _PhotonProcess_ object, _proc_, for photon map lookups

	PhotonProcess proc(nLookup, p);
	proc.photons =
		(ClosePhoton *)alloca(nLookup * sizeof(ClosePhoton));
	++lookups;  // NOBOOK

	map->Lookup(p, proc, maxDistSquared);

	// Accumulate light from nearby photons
	static StatsRatio foundRate("Volumetric Photon Map", "Photons found per lookup"); // NOBOOK
	foundRate.Add(proc.foundPhotons, 1); // NOBOOK
	float scale = 1.f / (float(nPaths) * maxDistCubed * M_PI * 4.f / 3.f);
	//float scale = 1.f / (maxDistCubed * M_PI * 4.f / 3.f);
	ClosePhoton *photons = proc.photons;
	int nFound = proc.foundPhotons;
	for(int i=0;i<nFound;i++)
		L += vr->p(p, photons[i].photon->wi, wo) * photons[i].photon->alpha * scale;

	return L;
}

void VolumePhotonIntegrator::WritePhotonsOutToMap(std::vector<Photon>* map)
{
	printf("\n");
	//printf("writing output file...\n");
	std::ofstream file;
	file.open("out.pbrt", std::ios::trunc);
	file << "AttributeBegin\n";
	file << "\tMaterial \"matte\" \"color Kd\" [1.0 1.0 1.0]\n";

	Point position = Point(0,0,0);
	for(int a=0;a<map->size();a++)
	{
		Photon photon = (*map)[a];
		if(isinf(photon.p.x) || isinf(photon.p.y) || isinf(photon.p.z))
		{
			continue;
		}
		Vector difference = photon.p - position;
		file << "Translate " << difference.x << " " << difference.y << " " << difference.z << "\n";
		file << "Shape \"sphere\" \"float radius\" [0.005]\n";
		position = photon.p;
	}
	file << "AttributeEnd\n";
	file << "WorldEnd\n";
	file.close();
	printf("file written\n");
}


extern "C" DLLEXPORT VolumeIntegrator *CreateVolumeIntegrator(const ParamSet &params) {
	float stepSize  = params.FindOneFloat("stepsize", 1.f);

	//use for dist^2 and dist^3
	float maxDist = params.FindOneFloat("maxdist", .1f);
	bool directPhotons = params.FindOneBool("directwithphotons", false);
	int nPhotons= params.FindOneInt("photons", 100000);
	float g = params.FindOneFloat("g", 0.f);
	int nUsed = params.FindOneInt("nused", 50);
	bool ignoreFirst = params.FindOneBool("ignorefirstlight", false);
	return new VolumePhotonIntegrator(stepSize, maxDist, directPhotons, nPhotons, g, nUsed, ignoreFirst);
}
